#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from unittest.mock import MagicMock

import pytest

from pyleco.json_utils.json_objects import Request, ErrorResponse, DataError
from pyleco.json_utils.errors import NODE_UNKNOWN, NOT_SIGNED_IN, DUPLICATE_NAME, RECEIVER_UNKNOWN
from pyleco.core import VERSION_B
from pyleco.core.message import Message, MessageTypes
from pyleco.core.leco_protocols import ExtendedComponentProtocol, Protocol, CoordinatorProtocol
from pyleco.utils.coordinator_utils import FakeMultiSocket, FakeNode
from pyleco.json_utils.rpc_generator import RPCGenerator
from pyleco.test import FakeContext
from pyleco.utils.events import SimpleEvent

from pyleco.coordinators.coordinator import Coordinator
from pyleco.coordinators import coordinator as coordinator_module  # type: ignore


@pytest.fixture
def coordinator() -> Coordinator:
    coordinator = Coordinator(namespace="N1", host="N1host", cleaning_interval=1e5,
                              context=FakeContext(),  # type: ignore
                              multi_socket=FakeMultiSocket()
                              )
    d = coordinator.directory
    d.add_component(b"send", b"321")
    d.add_component(b"rec", b"123")
    d.add_node_sender(FakeNode(), "N2host:12300", namespace=b"N2")
    d._nodes[b"N2"] = d._waiting_nodes.pop("N2host:12300")
    d._nodes[b"N2"].namespace = b"N2"
    d._waiting_nodes = {}
    d.add_node_receiver(b"n2", b"N2")
    n2 = coordinator.directory.get_node(b"N2")
    n2._messages_sent = []  # type: ignore # reset dealer sock._socket.
    n2.heartbeat = -1
    coordinator.sock._messages_sent = []  # type: ignore  # reset router sock._socket:
    return coordinator


def fake_perf_counter() -> float:
    return 0.


@pytest.fixture()
def fake_counting(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("pyleco.utils.coordinator_utils.perf_counter", fake_perf_counter)


cid = b"conversation_id;"


def fake_generate_cid() -> bytes:
    return cid


@pytest.fixture(autouse=True)
def fake_cid_generation(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr("pyleco.core.serialization.generate_conversation_id", fake_generate_cid)


class ExtendedCoordinator(CoordinatorProtocol, ExtendedComponentProtocol, Protocol):
    pass


class TestCoordinatorImplementsProtocol:
    protocol_methods = [m for m in dir(ExtendedCoordinator) if not m.startswith("_")]

    def static_test_methods_are_present(self):
        def testing(component: ExtendedCoordinator):
            pass
        testing(Coordinator())

    @pytest.fixture
    def component_methods(self, coordinator: Coordinator):
        response = coordinator.rpc.process_request(
            '{"id": 1, "method": "rpc.discover", "jsonrpc": "2.0"}')
        result = RPCGenerator().get_result_from_response(response)  # type: ignore
        return result.get('methods')

    @pytest.mark.parametrize("method", protocol_methods)
    def test_method_is_available(self, component_methods, method):
        for m in component_methods:
            if m.get('name') == method:
                return
        raise AssertionError(f"Method {method} is not available.")


class Test_coordinator_set_namespace_from_hostname:
    @pytest.fixture
    def namespace(self) -> bytes:
        coordinator = Coordinator(context=FakeContext())  # type: ignore
        return coordinator.namespace

    def test_namespace_is_bytes(self, namespace):
        assert isinstance(namespace, bytes)

    def test_namespace_without_periods(self, namespace):
        assert b"." not in namespace

    def test_namespace_from_hostname_with_periods(self, monkeypatch: pytest.MonkeyPatch):
        def fake_gethostname() -> str:
            return "hostname.domain.tld"
        monkeypatch.setattr(coordinator_module, "gethostname", fake_gethostname)
        coordinator = Coordinator(context=FakeContext())  # type: ignore
        assert coordinator.namespace == b"hostname"


def test_coordinator_set_namespace_bytes():
    coordinator = Coordinator(namespace=b"test", context=FakeContext())  # type: ignore
    assert coordinator.namespace == b"test"


def test_coordinator_set_namespace_invalid():
    with pytest.raises(ValueError, match="namespace"):
        Coordinator(namespace=1234, context=FakeContext())  # type: ignore


def test_set_address_from_hostname():
    coordinator = Coordinator(context=FakeContext())  # type: ignore
    assert isinstance(coordinator.address, str)


def test_set_address_manually():
    host = "host"
    coordinator = Coordinator(host=host, context=FakeContext())  # type: ignore
    assert coordinator.address == f"{host}:12300"


class TestClose:
    @pytest.fixture
    def coordinator_closed(self, coordinator: Coordinator):
        coordinator.shut_down = MagicMock()  # type: ignore[method-assign]
        coordinator.close()
        return coordinator

    def test_call_shutdown(self, coordinator_closed: Coordinator):
        coordinator_closed.shut_down.assert_called_once()  # type: ignore

    def test_close_socket(self, coordinator_closed: Coordinator):
        assert coordinator_closed.sock.closed is True


def test_context_manager_calls_close():
    with Coordinator(multi_socket=FakeMultiSocket()) as c:
        c.close = MagicMock()  # type: ignore[method-assign]
    c.close.assert_called_once()


class Test_clean_addresses:
    def test_expired_component(self, coordinator: Coordinator, fake_counting):
        coordinator.directory.get_components()[b"send"].heartbeat = -3.5
        coordinator.remove_expired_addresses(1)
        assert b"send" not in coordinator.directory.get_component_names()

    def test_expired_component_updates_directory(self, coordinator: Coordinator, fake_counting):
        coordinator.publish_directory_update = MagicMock()  # type: ignore
        coordinator.directory.get_components()[b"send"].heartbeat = -3.5
        coordinator.remove_expired_addresses(1)
        coordinator.publish_directory_update.assert_called()

    def test_warn_component(self, coordinator: Coordinator, fake_counting):
        # TODO implement heartbeat request
        coordinator.directory.get_components()[b"send"].heartbeat = -1.5
        coordinator.remove_expired_addresses(1)
        assert coordinator.sock._messages_sent == [(b"321", Message(  # type: ignore
            b"N1.send", b"N1.COORDINATOR",
            message_type=MessageTypes.JSON,
            data=Request(id=0, method="pong")))]

    def test_active_Component_remains_in_directory(self, coordinator: Coordinator, fake_counting):
        coordinator.directory.get_components()[b"send"].heartbeat = -0.5
        coordinator.remove_expired_addresses(1)
        assert coordinator.sock._messages_sent == []  # type: ignore
        assert b"send" in coordinator.directory.get_components()

    def test_expired_Coordinator(self, coordinator: Coordinator, fake_counting):
        coordinator.directory.get_node_ids()[b"n2"].heartbeat = -3.5
        coordinator.remove_expired_addresses(1)
        assert b"n2" not in coordinator.directory.get_node_ids()
        # further removal tests in :class:`Test_remove_coordinator`

    def test_warn_Coordinator(self, coordinator: Coordinator, fake_counting):
        coordinator.publish_directory_update = MagicMock()  # type: ignore
        coordinator.directory.get_node_ids()[b"n2"].heartbeat = -1.5
        coordinator.remove_expired_addresses(1)
        assert coordinator.directory.get_node_ids()[b"n2"]._messages_sent == [  # type: ignore
            Message(b"N2.COORDINATOR", b"N1.COORDINATOR",
                    message_type=MessageTypes.JSON,
                    data=Request(id=0, method="pong")),
        ]

    def test_active_Coordinator_remains_in_directory(self, coordinator: Coordinator, fake_counting):
        coordinator.directory.get_node_ids()[b"n2"].heartbeat = -0.5
        coordinator.remove_expired_addresses(1)
        assert b"n2" in coordinator.directory.get_node_ids()


def test_heartbeat_local(fake_counting, coordinator: Coordinator):
    coordinator.sock._messages_read = [  # type: ignore
        [b"321", Message(b"COORDINATOR", b"send")]]
    coordinator.read_and_route()
    assert coordinator.directory.get_components()[b"send"].heartbeat == 0


def test_routing_connects_to_coordinators(coordinator: Coordinator):
    event = SimpleEvent()
    event.set()
    coordinator.directory.add_node_sender = MagicMock()  # type: ignore
    coordinator.routing(["abc"], stop_event=event)
    coordinator.directory.add_node_sender.assert_called_once


@pytest.mark.parametrize("i, o", (
    ([b"321", VERSION_B, b"COORDINATOR", b"send", b";", b""], None),  # test heartbeat alone
    ([b"321", VERSION_B, b"rec", b"send", b";", b"1"],
     [b"123", VERSION_B, b"rec", b"send", b";", b"1"]),  # receiver known, sender known.
))
def test_routing_successful(coordinator: Coordinator, i, o):
    """Test whether some incoming message `i` is sent as `o`. Here: successful routing."""
    coordinator.sock._messages_read = [  # type: ignore
        (i[0], Message.from_frames(*i[1:]))]
    coordinator.read_and_route()
    if o is None:
        assert coordinator.sock._messages_sent == []  # type: ignore
    else:
        assert coordinator.sock._messages_sent == [  # type: ignore
            (o[0], Message.from_frames(*o[1:]))]


def test_reading_fails(coordinator: Coordinator, caplog: pytest.LogCaptureFixture):
    def read_message() -> tuple[bytes, Message]:
        return b"", Message.from_frames(*[b"frame 1", b"frame 2"])  # less frames than needed.
    coordinator.sock.read_message = read_message  # type: ignore
    coordinator.read_and_route()
    assert caplog.records[-1].msg == "Not enough frames read."


@pytest.mark.parametrize("i, o", (
    # receiver unknown, return to sender:
    ([b"321", VERSION_B, b"x", b"send", b"conversation_id;mid0", b""],
     [b"321", VERSION_B, b"send", b"N1.COORDINATOR", b"conversation_id;\x00\x00\x00\x01",
      ErrorResponse(id=None,
                    error=DataError.from_error(RECEIVER_UNKNOWN,
                                               "x")).model_dump_json().encode()]),
    # unknown receiver node:
    ([b"321", VERSION_B, b"N3.CB", b"N1.send", b"conversation_id;mid0"],
     [b"321", VERSION_B, b"N1.send", b"N1.COORDINATOR", b"conversation_id;\x00\x00\x00\x01",
      ErrorResponse(id=None,
                    error=DataError.from_error(NODE_UNKNOWN,
                                               "N3")).model_dump_json().encode()]),
    # sender (without namespace) did not sign in:
    ([b"1", VERSION_B, b"rec", b"unknownSender", b"conversation_id;mid0"],
     [b"1", VERSION_B, b"unknownSender", b"N1.COORDINATOR", b"conversation_id;\x00\x00\x00\x01",
      ErrorResponse(id=None, error=NOT_SIGNED_IN).model_dump_json().encode()]),
    # sender (with given Namespace) did not sign in:
    ([b"1", VERSION_B, b"rec", b"N1.unknownSender", b"conversation_id;mid0"],
     [b"1", VERSION_B, b"N1.unknownSender", b"N1.COORDINATOR", b"conversation_id;\x00\x00\x00\x01",
      ErrorResponse(id=None, error=NOT_SIGNED_IN).model_dump_json().encode()]),
    # unknown sender with a rogue node name:
    ([b"1", VERSION_B, b"rec", b"N2.unknownSender", b"conversation_id;mid0"],
     [b"1", VERSION_B, b"N2.unknownSender", b"N1.COORDINATOR", b"conversation_id;\x00\x00\x00\x01",
      ErrorResponse(id=None, error=NOT_SIGNED_IN).model_dump_json().encode()]),
))
def test_routing_error_messages(coordinator: Coordinator, i, o):
    """Test whether some incoming message `i` is sent as `o`. Here: Error messages."""
    coordinator.sock._messages_read = [  # type: ignore
        (i[0], Message.from_frames(*i[1:]))]
    coordinator.read_and_route()
    if o is None:
        assert coordinator.sock._messages_sent == []  # type: ignore
    else:
        assert coordinator.sock._messages_sent == [  # type: ignore
            (o[0], Message.from_frames(*o[1:]))]


def test_remote_routing(coordinator: Coordinator):
    coordinator.sock._messages_read = [  # type: ignore
        [b"321", Message(b"N2.CB", b"N1.send")]]
    coordinator.read_and_route()
    assert coordinator.directory.get_node(b"N2")._messages_sent == [  # type: ignore
        Message(b"N2.CB", b"N1.send")]


@pytest.mark.parametrize("sender", (b"N2.CB", b"N2.COORDINATOR"))
def test_remote_heartbeat(coordinator: Coordinator, fake_counting, sender):
    coordinator.sock._messages_read = [  # type: ignore
        [b"n2", Message(b"N3.CA", sender)]]
    assert coordinator.directory.get_node_ids()[b"n2"].heartbeat != 0
    coordinator.read_and_route()
    assert coordinator.directory.get_node_ids()[b"n2"].heartbeat == 0


class Test_handle_commands:
    class SpecialCoordinator(Coordinator):
        def handle_rpc_call(self, message: Message) -> None:
            self._rpc = message

    @pytest.fixture
    def coordinator_hc(self) -> Coordinator:
        return self.SpecialCoordinator(
            namespace="N1", host="N1host", cleaning_interval=1e5,
            context=FakeContext(),  # type: ignore
            multi_socket=FakeMultiSocket())

    def test_store_message(self, coordinator_hc: Coordinator):
        msg = Message(b"receiver", b"sender", header=b"header", data=b"data")
        coordinator_hc.handle_commands(b"identity", msg)
        assert coordinator_hc.current_message == msg

    def test_store_identity(self, coordinator_hc: Coordinator):
        msg = Message(b"receiver", b"sender", header=b"header", data=b"data")
        coordinator_hc.handle_commands(b"identity", msg)
        assert coordinator_hc.current_identity == b"identity"

    @pytest.mark.parametrize("identity, message", (
        (b"3", Message(b"", message_type=MessageTypes.JSON,
                       data={"jsonrpc": "2.0", "method": "some"})),
    ))
    def test_call_handle_rpc_call(self, coordinator_hc: Coordinator, identity, message):
        coordinator_hc.handle_commands(identity, message)
        assert coordinator_hc._rpc == message  # type: ignore

    def test_log_error_response(self, coordinator_hc: Coordinator):
        pass  # TODO

    def test_pass_at_null_result(self, coordinator_hc: Coordinator):
        coordinator_hc.handle_commands(b"",
                                       Message(b"",
                                               message_type=MessageTypes.JSON,
                                               data={"jsonrpc": "2.0", "result": None}))
        assert not hasattr(coordinator_hc, "_rpc")
        # assert no error log entry.  TODO

    def test_log_at_non_null_result(self, coordinator_hc: Coordinator,
                                    caplog: pytest.LogCaptureFixture):
        caplog.set_level(10)
        coordinator_hc.handle_commands(b"",
                                       Message(b"",
                                               message_type=MessageTypes.JSON,
                                               data={"jsonrpc": "2.0", "result": 5}))
        assert not hasattr(coordinator_hc, "_rpc")
        # assert no error log entry.  TODO
        caplog.records[-1].msg.startswith("Unexpected result")

    def test_pass_at_batch_of_null_results(self, coordinator_hc: Coordinator):
        coordinator_hc.handle_commands(b"",
                                       Message(b"",
                                               message_type=MessageTypes.JSON,
                                               data=[{"jsonrpc": "2.0", "result": None, "id": 1},
                                                     {"jsonrpc": "2.0", "result": None, "id": 2}]
                                               ))
        assert not hasattr(coordinator_hc, "_rpc")
        # assert no error log entry.  TODO

    def test_log_at_batch_of_non_null_results(self, coordinator_hc: Coordinator,
                                              caplog: pytest.LogCaptureFixture):
        caplog.set_level(10)
        coordinator_hc.handle_commands(b"",
                                       Message(b"",
                                               message_type=MessageTypes.JSON,
                                               data=[{"jsonrpc": "2.0", "result": None, "id": 1},
                                                     {"jsonrpc": "2.0", "result": 5, "id": 2}]
                                               ))
        assert not hasattr(coordinator_hc, "_rpc")
        caplog.records[-1].msg.startswith("Unexpected result")

    @pytest.mark.parametrize("data", (
            {"jsonrpc": "2.0", "no method": 7},
            ["jsonrpc", "2.0", "no method", 7],  # not a dict
    ))
    def test_invalid_json_does_not_raise_exception(self, coordinator_hc: Coordinator, data):
        coordinator_hc.handle_commands(b"",
                                       Message(receiver=b"COORDINATOR", sender=b"send",
                                               data=data, message_type=MessageTypes.JSON,))
        # assert that no error is raised

    def test_invalid_json_message_raises_log(self, coordinator_hc: Coordinator,
                                             caplog: pytest.LogCaptureFixture):
        data = "funny stuff"
        coordinator_hc.handle_commands(b"",
                                       Message(receiver=b"COORDINATOR", sender=b"send",
                                               data=data, message_type=MessageTypes.JSON,))
        assert caplog.records[-1].msg.startswith("Invalid JSON message")


class Test_sign_in:
    def test_signin(self, coordinator: Coordinator):
        coordinator.sock._messages_read = [  # type: ignore
            [b'cb', Message(b"COORDINATOR", b"CB",
                            data=Request(id=7, method="sign_in"),
                            message_type=MessageTypes.JSON,
                            conversation_id=cid,
                            )]]
        # read_and_route needs to start at routing, to check that the messages passes the heartbeats
        coordinator.read_and_route()
        assert coordinator.sock._messages_sent == [  # type: ignore
            (b"cb", Message(b"CB", b"N1.COORDINATOR",
                            conversation_id=cid,
                            message_type=MessageTypes.JSON,
                            data={"id": 7, "result": None, "jsonrpc": "2.0"}))]

    def test_signin_sends_directory_update(self, coordinator: Coordinator):
        coordinator.publish_directory_update = MagicMock()  # type: ignore
        coordinator.sock._messages_read = [  # type: ignore
            [b'cb', Message(b"COORDINATOR", b"CB", conversation_id=cid,
                            message_type=MessageTypes.JSON,
                            data={"jsonrpc": "2.0", "method": "sign_in", "id": 7},
                            )]]
        # read_and_route needs to start at routing, to check that the messages passes the heartbeats
        coordinator.read_and_route()
        coordinator.publish_directory_update.assert_any_call()

    def test_signin_rejected(self, coordinator: Coordinator):
        coordinator.sock._messages_read = [  # type: ignore
            [b'cb', Message(b"COORDINATOR", b"send", conversation_id=cid,
                            message_type=MessageTypes.JSON,
                            data={"id": 8, "method": "sign_in", "jsonrpc": "2.0"},
                            )]]
        coordinator.read_and_route()
        assert coordinator.sock._messages_sent == [(b"cb", Message(  # type: ignore
            b"send", b"N1.COORDINATOR",
            conversation_id=cid,
            message_type=MessageTypes.JSON,
            data={"id": None, "error": {"code": DUPLICATE_NAME.code,
                                        "message": DUPLICATE_NAME.message},
                  "jsonrpc": "2.0"}
        ))]


class Test_sign_out_successful:
    @pytest.fixture
    def coordinator_signed_out(self, coordinator: Coordinator):
        sign_out_message = Message(receiver=b"N1.COORDINATOR", sender=b"rec",
                                   message_type=MessageTypes.JSON,
                                   data={"jsonrpc": "2.0", "method": "sign_out", "id": 10})
        coordinator.publish_directory_update = MagicMock()  # type: ignore
        coordinator.sock._messages_read = [[b"123", sign_out_message]]  # type: ignore
        coordinator.read_and_route()
        return coordinator

    def test_address_cleared(self, coordinator_signed_out: Coordinator):
        assert b"rec" not in coordinator_signed_out.directory.get_components().keys()

    def test_acknowledgement_sent(self, coordinator_signed_out: Coordinator):
        assert coordinator_signed_out.sock._messages_sent == [  # type: ignore
            (b"123", Message(b"rec", b"N1.COORDINATOR",
                             message_type=MessageTypes.JSON,
                             data={"id": 10, "result": None, "jsonrpc": "2.0"}))]

    def test_directory_update_sent(self, coordinator_signed_out: Coordinator):
        coordinator_signed_out.publish_directory_update.assert_any_call()  # type: ignore

    def test_requires_new_sign_in(self, coordinator_signed_out):
        coordinator = coordinator_signed_out
        coordinator.sock._messages_sent = []  # type: ignore
        coordinator.sock._messages_read = [[b'123', Message(  # type: ignore
            b"N1.COORDINATOR", b"rec",
            message_type=MessageTypes.JSON,
            data={"jsonrpc": "2.0", "result": None, "id": 11})]]
        coordinator.read_and_route()
        assert coordinator.sock._messages_sent == [(b"123", Message(  # type: ignore
            b"rec", b"N1.COORDINATOR", message_type=MessageTypes.JSON,
            data=ErrorResponse(id=None, error=NOT_SIGNED_IN)))]


def test_sign_out_clears_address_explicit_namespace(coordinator: Coordinator):
    coordinator.sock._messages_read = [[b'123', Message(  # type: ignore
        b"N1.COORDINATOR", b"N1.rec", message_type=MessageTypes.JSON,
        data={"jsonrpc": "2.0", "method": "sign_out", "id": 10})]]
    coordinator.read_and_route()
    assert b"rec" not in coordinator.directory.get_components().keys()
    assert coordinator.sock._messages_sent == [  # type: ignore
        (b"123", Message(b"N1.rec", b"N1.COORDINATOR", message_type=MessageTypes.JSON,
                         data={"id": 10, "result": None, "jsonrpc": "2.0"}))]


def test_sign_out_of_not_signed_in_generates_acknowledgment_nonetheless(coordinator: Coordinator):
    coordinator.sock._messages_read = [[b'584', Message(  # type: ignore
        b"N1.COORDINATOR", b"rec584", message_type=MessageTypes.JSON,
        data={"jsonrpc": "2.0", "method": "sign_out", "id": 10})]]
    coordinator.read_and_route()
    assert coordinator.sock._messages_sent == [  # type: ignore
        (b"584", Message(b"rec584", b"N1.COORDINATOR", message_type=MessageTypes.JSON,
                         data={"id": 10, "result": None, "jsonrpc": "2.0"}))]


class Test_coordinator_sign_in:
    def test_co_signin_unknown_coordinator_successful(self, coordinator: Coordinator):
        """Test that an unknown Coordinator may sign in."""
        coordinator.sock._messages_read = [  # type: ignore
            [b'n3', Message(b"COORDINATOR", b"N3.COORDINATOR",
                            message_type=MessageTypes.JSON,
                            data={"jsonrpc": "2.0", "method": "coordinator_sign_in", "id": 15},
                            conversation_id=cid)]]
        coordinator.read_and_route()
        assert b'n3' in coordinator.directory.get_node_ids().keys()
        assert coordinator.sock._messages_sent == [  # type: ignore
            (b'n3', Message(b"COORDINATOR", b"N1.COORDINATOR",
                            message_type=MessageTypes.JSON,
                            conversation_id=cid, data={"id": 15, "result": None,
                                                       "jsonrpc": "2.0"}))]

    def test_co_signin_known_coordinator_successful(self, fake_counting, coordinator: Coordinator):
        """Test that a Coordinator may sign in as a response to N1's sign in."""

        coordinator.directory.add_node_sender(FakeNode(), "N3host:12345", namespace=b"N3")
        coordinator.directory.get_nodes()[b"N3"] = coordinator.directory._waiting_nodes.pop(
            "N3host:12345")
        coordinator.directory.get_nodes()[b"N3"].namespace = b"N3"

        coordinator.sock._messages_read = [  # type: ignore
            [b'n3', Message(b"COORDINATOR", b"N3.COORDINATOR",
                            conversation_id=cid,
                            message_type=MessageTypes.JSON,
                            data={"jsonrpc": "2.0", "method": "coordinator_sign_in", "id": 15},)]]
        coordinator.read_and_route()
        assert b'n3' in coordinator.directory.get_node_ids().keys()
        assert coordinator.sock._messages_sent == [(b'n3', Message(  # type: ignore
            b"COORDINATOR", b"N1.COORDINATOR", message_type=MessageTypes.JSON, conversation_id=cid,
            data={"id": 15, "result": None, "jsonrpc": "2.0"}))]

    @pytest.mark.xfail(True, reason="Additional error data is added")
    def test_co_signin_rejected(self, coordinator: Coordinator):
        """Coordinator sign in rejected due to already connected Coordinator."""
        coordinator.sock._messages_read = [  # type: ignore
            [b'n3', Message(b"COORDINATOR", b"N2.COORDINATOR",
                            data={"jsonrpc": "2.0", "method": "coordinator_sign_in", "id": 15},
                            message_type=MessageTypes.JSON,
                            conversation_id=cid)]]
        coordinator.read_and_route()
        assert coordinator.sock._messages_sent == [(b"n3", Message(  # type: ignore
            b"COORDINATOR", b"N1.COORDINATOR",
            data={"id": 15, "error": {"code": -32000, "message": "Server error",
                                      "data": "ValueError: Another Coordinator is known!"},
                  "jsonrpc": "2.0"},
            message_type=MessageTypes.JSON,
            conversation_id=cid))]

    def test_coordinator_sign_in_fails_at_duplicate_name(self, coordinator: Coordinator):
        coordinator.current_message = Message(
            b"COORDINATOR", b"N2.COORDINATOR",
            data={"jsonrpc": "2.0", "method": "coordinator_sign_in", "id": 15},
            message_type=MessageTypes.JSON,
            conversation_id=cid)
        coordinator.current_identity = b"n3"
        with pytest.raises(ValueError, match="Another Coordinator is known!"):
            coordinator.coordinator_sign_in()

    def test_co_signin_of_self_rejected(self, coordinator: Coordinator):
        """Coordinator sign in rejected because it is the same coordinator."""
        coordinator.sock._messages_read = [  # type: ignore
            [b'n3', Message(b"COORDINATOR", b"N1.COORDINATOR", conversation_id=cid,
                            message_type=MessageTypes.JSON,
                            data={"jsonrpc": "2.0", "method": "coordinator_sign_in", "id": 15})]]
        coordinator.read_and_route()
        assert coordinator.sock._messages_sent == [  # type: ignore
            (b'n3', Message(b"N1.COORDINATOR", b"N1.COORDINATOR", conversation_id=cid,
                            message_type=MessageTypes.JSON,
                            data=ErrorResponse(id=None, error=NOT_SIGNED_IN)))]


class Test_coordinator_sign_out:
    def test_co_signout_successful(self, coordinator: Coordinator):
        coordinator.sock._messages_read = [  # type: ignore
            [b'n2', Message(b"COORDINATOR", b"N2.COORDINATOR",
                            conversation_id=cid,
                            message_type=MessageTypes.JSON,
                            data={"id": 10, "method": "coordinator_sign_out", "jsonrpc": "2.0"})]]
        node = coordinator.directory.get_node(b"N2")
        coordinator.read_and_route()
        assert b"n2" not in coordinator.directory.get_node_ids()
        assert node._messages_sent == [Message(  # type: ignore
            b"N2.COORDINATOR", b"N1.COORDINATOR", conversation_id=cid,
            message_type=MessageTypes.JSON,
            data={"id": 100, "method": "coordinator_sign_out", "jsonrpc": "2.0"})]

    @pytest.mark.xfail(True, reason="Not yet defined.")
    def test_co_signout_rejected_due_to_different_identity(self, coordinator: Coordinator):
        """TODO TBD how to handle it"""
        coordinator.set_log_level("DEBUG")
        coordinator.sock._messages_read = [  # type: ignore
            [b'n4', Message(
                receiver=b"COORDINATOR", sender=b"N2.COORDINATOR", conversation_id=cid,
                message_type=MessageTypes.JSON,
                data={"id": 10, "method": "coordinator_sign_out", "jsonrpc": "2.0"})]]
        coordinator.read_and_route()
        assert coordinator.sock._messages_sent == [  # type: ignore
            (b"n4", Message(
                receiver=b"N2.COORDINATOR", sender=b"N1.COORDINATOR", conversation_id=cid,
                message_type=MessageTypes.JSON,
                data=ErrorResponse(id=None, error=NOT_SIGNED_IN)))]

    def test_co_signout_of_not_signed_in_coordinator(self, coordinator: Coordinator):
        """TODO TBD whether to reject or to ignore."""
        coordinator.sock._messages_read = [  # type: ignore
            (b"n4", Message(b"COORDINATOR", b"N4.COORDINATOR",
                            message_type=MessageTypes.JSON,
                            data={"id": 10, "method": "coordinator_sign_out", "jsonrpc": "2.0"}))]
        coordinator.read_and_route()
        assert coordinator.sock._messages_sent == []  # type: ignore


class Test_shutdown:
    @pytest.fixture
    def shutdown_coordinator(self, coordinator: Coordinator) -> Coordinator:
        self.n2 = coordinator.directory.get_node(b"N2")
        coordinator.stop_event = SimpleEvent()
        coordinator.shut_down()
        return coordinator

    def test_sign_out_message_to_other_coordinators_sent(self, shutdown_coordinator: Coordinator):
        assert self.n2._messages_sent == [  # type: ignore
            Message(b"N2.COORDINATOR", b"N1.COORDINATOR",
                    message_type=MessageTypes.JSON,
                    data={"id": 2, "method": "coordinator_sign_out", "jsonrpc": "2.0"})]

    def test_event_set(self, shutdown_coordinator: Coordinator):
        assert shutdown_coordinator.stop_event.is_set() is True


def test_send_nodes(coordinator: Coordinator):
    data = coordinator.send_nodes()
    assert data == {"N1": "N1host:12300", "N2": "N2host:12300"}


def test_send_local_components(coordinator: Coordinator):
    data = coordinator.send_local_components()
    assert data == ["send", "rec"]


def test_send_global_components(coordinator: Coordinator):
    # Arrange
    coordinator.global_directory[b"N5"] = ["some", "coordinator"]
    # Act
    data = coordinator.send_global_components()
    assert data == {
                 "N5": ["some", "coordinator"],
                 "N1": ["send", "rec"]}


class Test_add_nodes:
    def test_add_nodes(self, coordinator: Coordinator, fake_counting):
        coordinator.add_nodes({"N1": "N1host:12300", "N2": "wrong_host:-7", "N3": "N3host:12300"})
        assert coordinator.directory.get_node(b"N2").address == "N2host:12300"  # not changed
        assert "N3host:12300" in coordinator.directory._waiting_nodes.keys()  # newly created


class Test_record_components:
    def test_set(self, coordinator: Coordinator):
        coordinator.current_message = Message(b"", sender="N2.COORDINATOR")
        coordinator.record_components(["send", "rec"])
        assert coordinator.global_directory == {b"N2": ["send", "rec"]}


def test_publish_directory_updates(coordinator: Coordinator):
    # TODO TBD in LECO
    coordinator.publish_directory_update()
    assert coordinator.directory.get_node_ids()[b"n2"]._messages_sent == [  # type: ignore
        Message(
            b'N2.COORDINATOR', b'N1.COORDINATOR',
            message_type=MessageTypes.JSON,
            data=[
                {"id": 5, "method": "add_nodes",
                    "params": {"nodes": {"N1": "N1host:12300", "N2": "N2host:12300"}},
                    "jsonrpc": "2.0"},
                {"id": 6, "method": "record_components",
                    "params": {"components": ["send", "rec"]},
                    "jsonrpc": "2.0"}]
        ),
    ]



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
from unittest.mock import MagicMock
import time
from typing import Optional

import pytest

from pyleco.core import VERSION_B
from pyleco.core.message import Message, MessageTypes
from pyleco.core.leco_protocols import ExtendedComponentProtocol, LogLevels
from pyleco.core.internal_protocols import CommunicatorProtocol
from pyleco.core.serialization import serialize_data
from pyleco.test import FakeContext, FakePoller
from pyleco.json_utils.json_objects import Request, ResultResponse, ErrorResponse
from pyleco.json_utils.errors import JSONRPCError, INVALID_REQUEST, NOT_SIGNED_IN, DUPLICATE_NAME,\
    NODE_UNKNOWN, RECEIVER_UNKNOWN

from pyleco.utils.message_handler import MessageHandler, SimpleEvent


handler_name = "N1.handler"
remote_name = "remote"
cid = b"conversation_id;"
header = b"".join((cid, b"\x00" * 4))


def fake_generate_cid():
    return cid


@pytest.fixture()
def fake_cid_generation(monkeypatch):
    monkeypatch.setattr("pyleco.core.serialization.generate_conversation_id", fake_generate_cid)


@pytest.fixture()
def handler() -> MessageHandler:
    handler = MessageHandler(name=handler_name.split(".")[1], context=FakeContext())  # type: ignore
    handler.namespace = handler_name.split(".")[0]
    handler.stop_event = SimpleEvent()
    handler.timeout = 0.1
    return handler


class TestProtocolImplemented:
    protocol_methods = [m for m in dir(ExtendedComponentProtocol) if not m.startswith("_")]

    def static_test_methods_are_present(self):
        def testing(component: ExtendedComponentProtocol):
            pass
        testing(MessageHandler(name="test"))

        def test_internal_communicator_protocol(communicator: CommunicatorProtocol):
            pass
        test_internal_communicator_protocol(MessageHandler(name="test"))

    @pytest.fixture
    def component_methods(self, handler: MessageHandler):
        response = handler.rpc.process_request(
            '{"id": 1, "method": "rpc.discover", "jsonrpc": "2.0"}')
        result = handler.rpc_generator.get_result_from_response(response)  # type: ignore
        return result.get('methods')

    @pytest.mark.parametrize("method", protocol_methods)
    def test_method_is_available(self, component_methods, method):
        for m in component_methods:
            if m.get('name') == method:
                return
        raise AssertionError(f"Method {method} is not available.")


class Test_setup_logging:
    def test(self, handler: MessageHandler):
        logger = logging.getLogger("test")
        logger.addHandler(logging.NullHandler())
        handler.setup_logging(logger)
        assert len(logger.handlers) == 2
        assert handler.root_logger == logger
        assert handler.log == logging.getLogger("test.MessageHandler")


class Test_namespace_setter:
    def test_full_name_without_namespace(self, handler: MessageHandler):
        handler.namespace = None
        assert handler.full_name == "handler"

    @pytest.fixture
    def handler_ns(self, handler: MessageHandler) -> MessageHandler:
        handler.namespace = "xyz"
        return handler

    def test_namespace(self, handler_ns: MessageHandler):
        assert handler_ns.namespace == "xyz"

    def test_full_name(self, handler_ns: MessageHandler):
        assert handler_ns.full_name == "xyz.handler"

    def test_rpc_title(self, handler_ns: MessageHandler):
        assert handler_ns.rpc.title == "xyz.handler"

    def test_log_handler(self, handler_ns: MessageHandler):
        assert handler_ns.log_handler.full_name == "xyz.handler"


class Test_sign_in:
    def test_sign_in_successful(self, handler: MessageHandler, fake_cid_generation):
        message = Message(receiver=b"N3.handler", sender=b"N3.COORDINATOR",
                          conversation_id=cid,
                          message_type=MessageTypes.JSON,
                          data={
                              "id": 0, "result": None, "jsonrpc": "2.0",
                              })
        handler.socket._r = [message.to_frames()]  # type: ignore
        handler.namespace = None
        handler.sign_in()
        assert handler.namespace == "N3"

    def test_not_valid_message(self, handler: MessageHandler, caplog: pytest.LogCaptureFixture,
                               fake_cid_generation):
        message = Message("handler", "COORDINATOR", data=b"[]", conversation_id=cid)
        handler.socket._r = [message.to_frames()]  # type: ignore
        handler.sign_in()
        caplog.records[-1].msg.startswith("Not json message received:")

    def test_duplicate_name(self, handler: MessageHandler, caplog: pytest.LogCaptureFixture,
                            fake_cid_generation):
        handler.namespace = None
        message = Message("handler", "N3.COORDINATOR", message_type=MessageTypes.JSON,
                          data=ErrorResponse(id=5, error=DUPLICATE_NAME),
                          conversation_id=cid)
        handler.socket._r = [message.to_frames()]  # type: ignore
        handler.sign_in()
        assert handler.namespace is None
        assert caplog.records[-1].msg == "Sign in failed, the name is already used."

    def test_handle_unknown_error(self, handler: MessageHandler, caplog: pytest.LogCaptureFixture,
                                  fake_cid_generation):
        handler.namespace = None
        message = Message("handler", "N3.COORDINATOR", message_type=MessageTypes.JSON, data={
            "jsonrpc": "2.0", "error": {'code': 123545, "message": "error_msg"}, "id": 5
        }, conversation_id=cid)
        handler.socket._r = [message.to_frames()]  # type: ignore
        handler.sign_in()
        assert handler.namespace is None
        assert caplog.records[-1].msg.startswith("Sign in failed, unknown error")

    def test_handle_request_message(self, handler: MessageHandler, caplog: pytest.LogCaptureFixture,
                                    fake_cid_generation
                                    ):
        """Handle a message without result or error."""
        handler.namespace = None
        message = Message("handler", "N3.COORDINATOR", message_type=MessageTypes.JSON, data={
            "jsonrpc": "2.0", "id": 5, "method": "some_method",
        }, conversation_id=cid)
        handler.socket._r = [message.to_frames()]  # type: ignore
        handler.sign_in()
        assert handler.namespace is None
        assert caplog.records[-1].msg.startswith("Sign in failed, unknown error")

    def test_log_timeout_error(self, handler: MessageHandler, caplog: pytest.LogCaptureFixture):
        handler.sign_in()
        assert caplog.records[-1].msg.startswith("Signing in timed out.")


class Test_finish_sign_in:
    @pytest.fixture
    def handler_fsi(self, handler: MessageHandler, caplog: pytest.LogCaptureFixture):
        caplog.set_level(logging.INFO)
        handler.finish_sign_in(response_message=Message(
            b"handler", b"N5.COORDINATOR",
            message_type=MessageTypes.JSON,
            data={"id": 10, "result": None, "jsonrpc": "2.0"}))
        return handler

    def test_namespace(self, handler_fsi: MessageHandler):
        assert handler_fsi.namespace == "N5"

    def test_full_name(self, handler_fsi: MessageHandler):
        assert handler_fsi.full_name == "N5.handler"

    def test_log_message(self, handler_fsi: MessageHandler, caplog: pytest.LogCaptureFixture):
        assert caplog.get_records("setup")[-1].message == ("Signed in to Node 'N5'.")


def test_sign_out_fail(handler: MessageHandler, caplog: pytest.LogCaptureFixture,
                       fake_cid_generation):
    handler.namespace = "N3"
    message = Message("handler", "N3.COORDINATOR", message_type=MessageTypes.JSON, data={
        "jsonrpc": "2.0", "error": {"code": 12345}, "id": 1,
    }, conversation_id=cid)
    handler.socket._r = [message.to_frames()]  # type: ignore
    handler.sign_out()
    assert handler.namespace is not None
    assert caplog.messages[-1].startswith("Signing out failed")


def test_sign_out_success(handler: MessageHandler, fake_cid_generation):
    handler.namespace = "N3"
    message = Message("handler", "N3.COORDINATOR", message_type=MessageTypes.JSON, data={
        "jsonrpc": "2.0", "result": None, "id": 1,
    }, conversation_id=cid)
    handler.socket._r = [message.to_frames()]  # type: ignore
    handler.sign_out()
    assert handler.namespace is None


def test_finish_sign_out(handler: MessageHandler):
    handler.finish_sign_out()
    assert handler.namespace is None
    assert handler.full_name == "handler"


# test communication
def test_send(handler: MessageHandler):
    handler.send("N2.CB", conversation_id=cid, message_id=b"sen", data=[["TEST"]],
                 message_type=MessageTypes.JSON)
    assert handler.socket._s == [[VERSION_B, b"N2.CB", b"N1.handler", b"conversation_id;sen\x01",
                                  b'[["TEST"]]']]


def test_send_with_sender(handler: MessageHandler):
    handler.send("N2.CB", sender="sender", conversation_id=cid, message_id=b"sen",
                 data=[["TEST"]],
                 message_type=MessageTypes.JSON)
    assert handler.socket._s == [[VERSION_B, b"N2.CB", b"sender", b"conversation_id;sen\x01",
                                  b'[["TEST"]]']]


def test_send_message_raises_error(handler: MessageHandler, caplog: pytest.LogCaptureFixture):
    handler.send(receiver=remote_name, header=b"header", conversation_id=b"12345")
    assert caplog.messages[-1].startswith("Composing message with")


def test_heartbeat(handler: MessageHandler, fake_cid_generation):
    handler.heartbeat()
    assert handler.socket._s == [[VERSION_B, b"COORDINATOR", b"N1.handler", header]]


class Test_read_message:
    m1 = Message(receiver=handler_name, sender="xy")  # some message
    mr = Message(receiver=handler_name, sender="xy", conversation_id=cid)  # requested message
    m2 = Message(receiver=handler_name, sender="xy")  # another message

    conf: list[tuple[list[Message], list[Message], Optional[bytes], list[Message], list[Message],
                     str]] = [
        # socket_in, buffer_in, cid, socket_out, buffer_out, test_id
        # find first not requested message
        ([m1], [], None, [], [], "return first message from socket"),
        ([m2], [m1], None, [m2], [], "return first message from buffer, not socket"),
        ([m1], [mr], None, [], [mr], "ignore requested message in buffer"),
        ([mr, m1], [], None, [], [mr], "ignore requested message in socket"),
        # find requested message
        ([mr], [], cid, [], [], "return specific message from socket"),
        ([m2], [mr], cid, [m2], [], "return specific message from buffer"),
        ([mr], [m2], cid, [], [m2], "return specific message from socket although filled buffer"),
        ([m2, mr, m1], [], cid, [m1], [m2], "find specific message in socket"),
        ([], [m2, mr, m1], cid, [], [m2, m1], "find specific message in buffer"),
    ]
    ids = [test[-1] for test in conf]

    def test_return_message_from_socket(self, handler: MessageHandler):
        handler.socket._r = [self.m1.to_frames()]  # type: ignore
        assert handler.read_message() == self.m1

    def test_return_message_from_buffer(self, handler: MessageHandler):
        handler.message_buffer.add_message(self.m1)
        assert handler.read_message() == self.m1
        # assert that no error is raised

    def test_cid_not_longer_in_requested_ids(self, handler: MessageHandler):
        handler.message_buffer.add_conversation_id(cid)
        handler.message_buffer.add_message(self.mr)
        handler.read_message(conversation_id=cid)
        assert handler.message_buffer.is_conversation_id_requested(cid) is False

    @pytest.mark.parametrize("test", conf, ids=ids)
    def test_return_correct_message(self,
                                    test: tuple[list[Message], list, Optional[bytes]],
                                    handler: MessageHandler):
        socket, buffer, cid0, *_ = test
        handler.socket._r = [m.to_frames() for m in socket]  # type: ignore
        for m in buffer:
            handler.message_buffer.add_message(m)
        handler.message_buffer.add_conversation_id(cid)
        # act and assert
        assert handler.read_message(conversation_id=cid0) == self.m1 if cid is None else self.mr

    @pytest.mark.parametrize("test", conf, ids=ids)
    def test_correct_buffer_socket(self, test, handler: MessageHandler):
        socket_in, buffer_in, cid0, socket_out, buffer_out, *_ = test
        handler.socket._r = [m.to_frames() for m in socket_in]  # type: ignore
        for m in buffer_in:
            handler.message_buffer.add_message(m)
        handler.message_buffer.add_conversation_id(cid)
        # act
        handler.read_message(conversation_id=cid0)
        assert handler.socket._r == [m.to_frames() for m in socket_out]  # type: ignore
        assert handler.message_buffer._messages == buffer_out

    def test_timeout_zero_works(self, handler: MessageHandler):
        handler.socket._r = [self.m1.to_frames()]  # type: ignore
        handler.read_message(timeout=0)
        # assert that no error is raised

    def test_timeout_error(self, handler: MessageHandler):
        def waiting(*args, **kwargs):
            time.sleep(.1)
            return self.m1
        handler._read_socket_message = waiting  # type: ignore[assignment]
        with pytest.raises(TimeoutError):
            handler.read_message(conversation_id=cid, timeout=0)


class Test_ask_message:
    expected_sent = Message(remote_name, sender=handler_name, conversation_id=cid)
    expected_response = Message(handler_name, sender=remote_name, conversation_id=cid)

    @pytest.fixture
    def handler_asked(self, handler: MessageHandler):
        handler.socket._r = [self.expected_response.to_frames()]  # type: ignore
        self.response = handler.ask_message(message=self.expected_sent)
        return handler

    def test_sent_expected(self, handler_asked: MessageHandler):
        assert handler_asked.socket._s == [self.expected_sent.to_frames()]

    def test_expected_response(self, handler_asked):
        assert self.expected_response == self.response

    def test_no_cid_in_requested_cids_list(self, handler_asked: MessageHandler):
        assert handler_asked.message_buffer.is_conversation_id_requested(cid) is False


class Test_read_and_handle_message:
    def test_handle_message_handles_no_new_socket_message(self, handler: MessageHandler):
        """Test, that the message handler does not raise an error without a new socket message."""
        handler.message_buffer.add_conversation_id(cid)
        handler.socket._r = [  # type: ignore
            Message(receiver=handler_name, sender=remote_name, conversation_id=cid).to_frames()]
        # act
        handler.read_and_handle_message()
        # assert that no error is raised.

    def test_handle_message_ignores_heartbeats(self, handler: MessageHandler):
        handler.handle_message = MagicMock()  # type: ignore
        # empty message of heartbeat
        handler.socket._r = [[VERSION_B, b"N1.handler", b"whatever", b";"]]  # type: ignore
        handler.read_and_handle_message()
        handler.handle_message.assert_not_called()

    @pytest.mark.parametrize("i, out", (
        (  # shutdown
         [VERSION_B, b"N1.handler", b"N1.CB", b"conversation_id;mid" + bytes((MessageTypes.JSON,)),
          serialize_data({"id": 5, "method": "shut_down", "jsonrpc": "2.0"})],
         [VERSION_B, b"N1.CB", b"N1.handler", b"conversation_id;\x00\x00\x00\x00",
          serialize_data({"id": 5, "result": None, "jsonrpc": "2.0"})]),
        (  # pong
         Message("N1.handler", "N1.COORDINATOR", conversation_id=cid,
                 message_type=MessageTypes.JSON, data=Request(id=2, method="pong")
                 ).to_frames(),
         Message("N1.COORDINATOR", "N1.handler", conversation_id=cid,
                 message_type=MessageTypes.JSON,
                 data=ResultResponse(id=2, result=None)).to_frames()),
    ))
    def test_read_and_handle_message(self, handler: MessageHandler,
                                     i: list[bytes], out: list[bytes]):
        handler.socket._r = [i]  # type: ignore
        handler.read_and_handle_message()
        for j in range(len(out)):
            if j == 3:
                continue  # reply adds timestamp
            assert handler.socket._s[0][j] == out[j]  # type: ignore

    def test_handle_not_signed_in_message(self, handler: MessageHandler):
        handler.sign_in = MagicMock()  # type: ignore
        handler.socket._r = [Message(receiver="handler", sender="N1.COORDINATOR",  # type: ignore
                                     message_type=MessageTypes.JSON,
                                     data=ErrorResponse(id=5, error=NOT_SIGNED_IN),
                                     ).to_frames()]
        handler.read_and_handle_message()
        assert handler.namespace is None
        handler.sign_in.assert_called_once()
        assert handler.full_name == "handler"

    def test_handle_node_unknown_message(self, handler: MessageHandler):
        error = Message("N1.handler", "N1.COORDINATOR", message_type=MessageTypes.JSON,
                        data=ErrorResponse(id=None, error=NODE_UNKNOWN))
        handler.message_buffer.add_message(error)
        handler.read_and_handle_message()
        # assert that no error is raised and that no message is sent
        assert handler.socket._s == []

    def test_handle_receiver_unknown_message(self, handler: MessageHandler):
        error = Message("N1.handler", "N1.COORDINATOR", message_type=MessageTypes.JSON,
                        data=ErrorResponse(id=None, error=RECEIVER_UNKNOWN))
        handler.message_buffer.add_message(error)
        handler.read_and_handle_message()
        # assert that no error is raised and that no message is sent
        assert handler.socket._s == []

    def test_handle_ACK_does_not_change_Namespace(self, handler: MessageHandler):
        """Test that an ACK does not change the Namespace, if it is already set."""
        handler.socket._r = [Message(b"N3.handler", b"N3.COORDINATOR",  # type: ignore
                                     message_type=MessageTypes.JSON,
                                     data={"id": 3, "result": None, "jsonrpc": "2.0"}).to_frames()]
        handler.namespace = "N1"
        handler.read_and_handle_message()
        assert handler.namespace == "N1"

    def test_handle_invalid_json_message(self, handler: MessageHandler,
                                         caplog: pytest.LogCaptureFixture):
        """An invalid message should not cause the message handler to crash."""
        handler.socket._r = [Message(b"N3.handler", b"N3.COORDINATOR",  # type: ignore
                                     message_type=MessageTypes.JSON,
                                     data={"without": "method..."}).to_frames()]
        handler.read_and_handle_message()
        assert caplog.records[-1].msg.startswith("Invalid JSON message")

    def test_handle_corrupted_message(self, handler: MessageHandler,
                                      caplog: pytest.LogCaptureFixture):
        """An invalid message should not cause the message handler to crash."""
        handler.socket._r = [Message(b"N3.handler", b"N3.COORDINATOR",  # type: ignore
                                     message_type=MessageTypes.JSON,
                                     data=[]).to_frames()]
        handler.read_and_handle_message()
        assert caplog.records[-1].msg.startswith("Invalid JSON message")

    def test_handle_undecodable_message(self, handler: MessageHandler,
                                        caplog: pytest.LogCaptureFixture):
        """An invalid message should not cause the message handler to crash."""
        message = Message(
            b"N3.handler",
            b"N3.COORDINATOR",
            message_type=MessageTypes.JSON,
            additional_payload=[b"()"],
        )
        handler.socket._r = [message.to_frames()]  # type: ignore
        handler.read_and_handle_message()
        assert caplog.records[-1].msg.startswith("Could not decode")


def test_handle_unknown_message_type(handler: MessageHandler, caplog: pytest.LogCaptureFixture):
    message = Message(handler_name, sender="sender", message_type=255)
    handler.handle_message(message=message)
    assert caplog.records[-1].message.startswith("Message from b'sender'")


class Test_process_json_message:
    def test_handle_rpc_request(self, handler: MessageHandler):
        message = Message(receiver=handler_name, sender=remote_name,
                          data=Request(id=5, method="pong"),
                          conversation_id=cid, message_type=MessageTypes.JSON)
        response = Message(receiver=remote_name,
                           data=ResultResponse(id=5, result=None),
                           conversation_id=cid, message_type=MessageTypes.JSON)
        result = handler.process_json_message(message=message)
        assert result == response

    def test_handle_json_not_request(self, handler: MessageHandler):
        """Test, that a json message, which is not a request, is handled appropriately."""
        data = ResultResponse(id=5, result=None)  # some json, which is not a request.
        message = Message(receiver=handler_name, sender=remote_name,
                          data=data,
                          conversation_id=cid, message_type=MessageTypes.JSON)
        result = handler.process_json_message(message=message)
        assert result.receiver == remote_name.encode()
        assert result.conversation_id == cid
        assert result.header_elements.message_type == MessageTypes.JSON
        with pytest.raises(JSONRPCError) as exc_info:
            handler.rpc_generator.get_result_from_response(result.data)  # type: ignore
        error = exc_info.value.rpc_error
        assert error.code == INVALID_REQUEST.code
        assert error.message == INVALID_REQUEST.message


class Test_process_json_message_with_created_binary:
    payload_in: list[bytes]
    payload_out: list[bytes]

    @pytest.fixture(
        params=(
            # normally created binary method
            {"method": "do_binary", "params": [5]},  # with a list
            {"method": "do_binary", "params": {"data": 5}},  # a dictionary
            # manually created binary method
            {"method": "do_binary_manually", "params": [5]},
            {"method": "do_binary_manually", "params": {"data": 5}},
        ),
        ids=(
            "created, list",
            "created, dict",
            "manual, list",
            "manual, dict",
        ),
    )
    def data(self, request):
        """Create a request with a list and a dict of other parameters."""
        d = {"jsonrpc": "2.0", "id": 8}
        d.update(request.param)
        return d

    @pytest.fixture
    def handler_b(self, handler: MessageHandler):
        test_class = self
        class SpecialHandler(MessageHandler):
            def do_binary_manually(self, data: int) -> int:
                test_class.payload_in = self.current_message.payload[1:]
                self.additional_response_payload = test_class.payload_out
                return data

            def do_binary(
                self, data: int, additional_payload: Optional[list[bytes]] = None
            ) -> tuple[int, list[bytes]]:
                test_class.payload_in = additional_payload  # type: ignore
                return data, test_class.payload_out

        handler = SpecialHandler(name=handler_name.split(".")[1], context=FakeContext())  # type: ignore
        handler.namespace = handler_name.split(".")[0]
        handler.stop_event = SimpleEvent()
        handler.timeout = 0.1

        handler.register_rpc_method(handler.do_binary_manually)
        handler.register_binary_rpc_method(
            handler.do_binary, accept_binary_input=True, return_binary_output=True
        )
        return handler

    def test_message_stored(self, handler_b: MessageHandler, data):
        m_in = Message("abc", data=data, message_type=MessageTypes.JSON)
        handler_b.process_json_message(m_in)
        assert handler_b.current_message == m_in

    def test_empty_additional_payload(self, handler_b: MessageHandler, data):
        m_in = Message("abc", data=data, message_type=MessageTypes.JSON)
        handler_b.process_json_message(m_in)
        assert handler_b.additional_response_payload is None

    def test_binary_payload_available(self, handler_b: MessageHandler, data):
        m_in = Message(
            "abc", data=data, message_type=MessageTypes.JSON, additional_payload=[b"def"]
        )
        self.payload_out = []
        handler_b.process_json_message(m_in)
        assert self.payload_in == [b"def"]

    def test_binary_payload_sent(self, handler_b: MessageHandler, data):
        m_in = Message("abc", data=data, message_type=MessageTypes.JSON)
        self.payload_out = [b"ghi"]
        response = handler_b.process_json_message(m_in)
        assert response.payload[1:] == [b"ghi"]
        assert response.data == {"jsonrpc": "2.0", "id": 8, "result": 5}


def test_handle_binary_return_value(handler: MessageHandler):
    payload = [b"abc", b"def"]
    result = handler._handle_binary_return_value((None, payload))
    assert result is None
    assert handler.additional_response_payload == payload


class Test_generate_binary_method:
    @pytest.fixture
    def binary_method(self):
        def binary_method(index: int, additional_payload: list[bytes]) -> tuple[None, list[bytes]]:
            """Docstring of binary method."""
            return None, [additional_payload[index]]
        return binary_method

    @pytest.fixture(params=(True, False))
    def modified_binary_method(self, handler: MessageHandler, binary_method, request):
        handler.current_message = Message(
            "rec", "send", data=b"", additional_payload=[b"0", b"1", b"2", b"3"]
        )
        self._accept_binary_input = request.param
        mod = handler._generate_binary_capable_method(
            binary_method, accept_binary_input=self._accept_binary_input, return_binary_output=True
        )
        self.handler = handler
        return mod

    def test_name(self, binary_method, modified_binary_method):
        assert modified_binary_method.__name__ == binary_method.__name__

    def test_docstring(self, modified_binary_method, binary_method):
        doc_addition = (
            "(binary input output method)"
            if self._accept_binary_input
            else "(binary output method)"
        )
        assert modified_binary_method.__doc__ == binary_method.__doc__ + "\n" + doc_addition

    @pytest.mark.parametrize(
        "input, output, string",
        (
            (False, False, "(binary method)"),
            (True, False, "(binary input method)"),
            (False, True, "(binary output method)"),
            (True, True, "(binary input output method)"),
        ),
    )
    def test_docstring_without_original_docstring(
        self, handler: MessageHandler, input, output, string
    ):
        def binary_method(additional_payload):
            return 7
        mod = handler._generate_binary_capable_method(
            binary_method, accept_binary_input=input, return_binary_output=output
        )
        assert mod.__doc__ == string

    def test_annotation(self, modified_binary_method, binary_method):
        assert modified_binary_method.__annotations__ == binary_method.__annotations__

    def test_functionality_kwargs(self, modified_binary_method):
        if self._accept_binary_input:
            assert modified_binary_method(index=1) is None
        else:
            assert (
                modified_binary_method(index=1, additional_payload=[b"0", b"1", b"2", b"3"]) is None
            )
        assert self.handler.additional_response_payload == [b"1"]

    def test_functionality_args(self, modified_binary_method):
        if self._accept_binary_input:
            assert modified_binary_method(1) is None
        else:
            assert modified_binary_method(1, [b"0", b"1", b"2", b"3"]) is None
        assert self.handler.additional_response_payload == [b"1"]

    def test_binary_input_from_message(self, handler: MessageHandler):
        handler.current_message = Message("rec", "send", data=b"", additional_payload=[b"0"])

        def binary_method(additional_payload = None):
            return 7
        mod = handler._generate_binary_capable_method(
            binary_method, accept_binary_input=True, return_binary_output=False
        )
        assert mod() == 7


class Test_listen:
    @pytest.fixture
    def handler_l(self, handler: MessageHandler, fake_cid_generation):
        event = SimpleEvent()
        event.set()
        handler.socket._r = [  # type: ignore
            Message("handler", "N1.COORDINATOR", message_type=MessageTypes.JSON,
                    conversation_id=cid,
                    data={"id": 2, "result": None, "jsonrpc": "2.0"}).to_frames()]
        handler.listen(stop_event=event)
        return handler

    def test_messages_are_sent(self, handler_l: MessageHandler):
        assert handler_l.socket._s == [
            Message("COORDINATOR", "N1.handler", conversation_id=cid,
                    message_type=MessageTypes.JSON,
                    data={"id": 1, "method": "sign_in", "jsonrpc": "2.0"}).to_frames(),
            Message("COORDINATOR", "N1.handler", conversation_id=cid,
                    message_type=MessageTypes.JSON,
                    data={"id": 2, "method": "sign_out", "jsonrpc": "2.0"}).to_frames(),
        ]

    def test_next_beat(self, handler_l: MessageHandler):
        assert handler_l.next_beat > 0

    def test_loop_element_changes_heartbeat(self, handler_l: MessageHandler):
        handler_l.next_beat = 0
        # Act
        handler_l._listen_loop_element(poller=FakePoller(), waiting_time=0)  # type: ignore
        assert handler_l.next_beat > 0

    def test_loop_element_does_not_change_heartbeat_if_short(self, handler_l: MessageHandler):
        handler_l.next_beat = float("inf")
        # Act
        handler_l._listen_loop_element(poller=FakePoller(), waiting_time=0)  # type: ignore
        assert handler_l.next_beat == float("inf")

    def test_KeyboardInterrupt_in_loop(self, handler: MessageHandler):
        def raise_error(poller, waiting_time):
            raise KeyboardInterrupt
        handler.sign_in = MagicMock()  # type: ignore[method-assign]
        handler._listen_loop_element = raise_error  # type: ignore
        handler.listen()
        # assert that no error is raised and that the test does not hang


def test_listen_loop_element(handler: MessageHandler):
    poller = FakePoller()
    poller.register(handler.socket)
    handler.socket._r = [  # type: ignore
        Message("Test", "COORDINATOR").to_frames()
    ]
    socks = handler._listen_loop_element(poller, 0)  # type: ignore
    assert socks == {}


class Test_listen_close:
    @pytest.fixture
    def handler_lc(self, handler: MessageHandler):
        handler._listen_close(0)
        return handler

    def test_sign_out_sent(self, handler_lc: MessageHandler):
        sent = Message.from_frames(*handler_lc.socket._s[-1])  # type: ignore
        assert handler_lc.socket._s == [Message("COORDINATOR", "N1.handler",
                                                conversation_id=sent.conversation_id,
                                                message_type=MessageTypes.JSON,
                                                data={
                                                    "id": 1, "method": "sign_out", "jsonrpc": "2.0",
                                                    },
                                                ).to_frames()]

    def test_warning_log_written(self, handler_lc: MessageHandler,
                                 caplog: pytest.LogCaptureFixture):
        assert caplog.get_records("setup")[-1].message == "Waiting for sign out response timed out."


def test_set_log_level(handler: MessageHandler):
    handler.set_log_level(LogLevels.ERROR)
    assert handler.root_logger.level == 40  # logging.ERROR


def test_shutdown(handler: MessageHandler):
    handler.stop_event = SimpleEvent()
    handler.shut_down()
    assert handler.stop_event.is_set() is True



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import logging

import pytest

from pyleco.utils.parser import parser, parse_command_line_parameters


class Test_parse_command_line_parameters:
    @pytest.fixture
    def parsed_kwargs(self):
        self.parser = parser
        return parse_command_line_parameters(parser=self.parser,
                                             arguments=["-v", "--name", "name_value", "-v"],
                                             parser_description="Some description",
                                             logging_default=logging.WARNING
                                             )

    def test_parser_description(self, parsed_kwargs):
        assert self.parser.description == "Some description"

    def test_kwargs(self, parsed_kwargs):
        assert parsed_kwargs == {'name': "name_value"}

    def test_logging_level(self, parsed_kwargs):
        assert logging.getLogger("__main__").level == logging.DEBUG



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.json_utils import json_objects


def test_request():
    request = json_objects.Request(id=5, method="call")
    assert request.model_dump() == {
        "id": 5,
        "jsonrpc": "2.0",
        "method": "call",
    }


def test_request_with_params():
    request = json_objects.ParamsRequest(id=5, method="call", params=[1, 5])
    assert request.model_dump() == {
        "id": 5,
        "jsonrpc": "2.0",
        "method": "call",
        "params": [1, 5],
    }


def test_notification():
    request = json_objects.Notification(method="call")
    assert request.model_dump() == {
        "jsonrpc": "2.0",
        "method": "call",
    }


def test_notification_with_params():
    request = json_objects.ParamsNotification(method="call", params=[1, 5])
    assert request.model_dump() == {
        "jsonrpc": "2.0",
        "method": "call",
        "params": [1, 5],
    }


def test_result():
    result = json_objects.ResultResponse(id=5, result=7)
    assert result.model_dump() == {
        "id": 5,
        "jsonrpc": "2.0",
        "result": 7,
    }


def test_error_with_data():
    """Test that the Error object is json serializable."""
    data_error = json_objects.DataError(code=5, message="whatever", data="abc")
    error_response = json_objects.ErrorResponse(id=7, error=data_error)
    assert error_response.model_dump_json() == '{"id":7,"error":{"code":5,"message":"whatever","data":"abc"},"jsonrpc":"2.0"}'  # noqa


def test_generate_data_error_from_error():
    error = json_objects.Error(code=5, message="abc")
    data_error = json_objects.DataError.from_error(error, "data")
    assert data_error.code == error.code
    assert data_error.message == error.message
    assert data_error.data == "data"


class Test_BatchObject:
    element = json_objects.Request(5, "start")

    @pytest.fixture
    def batch_obj(self):
        return json_objects.BatchObject([self.element])

    def test_init_with_value(self):
        obj = json_objects.BatchObject([self.element])
        assert obj == [self.element]

    def test_init_with_values(self):
        obj = json_objects.BatchObject([self.element, self.element])
        assert obj == [self.element, self.element]

    def test_bool_value_with_element(self):
        obj = json_objects.BatchObject([self.element])
        assert bool(obj) is True

    def test_bool_value_without_element(self):
        obj = json_objects.BatchObject()
        assert bool(obj) is False

    def test_append(self, batch_obj: json_objects.BatchObject):
        el2 = json_objects.Request(5, "start")
        batch_obj.append(el2)
        assert batch_obj[-1] == el2

    def test_model_dump(self, batch_obj: json_objects.BatchObject):
        assert batch_obj.model_dump() == [self.element.model_dump()]

    def test_model_dump_json(self, batch_obj: json_objects.BatchObject):
        result = '[{"id":5,"method":"start","jsonrpc":"2.0"}]'
        assert batch_obj.model_dump_json() == result



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
from __future__ import annotations

from typing import Any, Callable, Generic, Optional, Sequence, TypeVar, Union
from warnings import warn

import zmq

from ..utils.message_handler import MessageHandler
from ..utils.data_publisher import DataPublisher
from ..utils.timers import RepeatingTimer


Device = TypeVar("Device")


class Actor(MessageHandler, Generic[Device]):
    """Control an instrument listening to zmq messages and regularly read some values.

    .. code::

        a = Actor("testing", TestClass)
        # define some function `readout(device: Device, publisher: DataPublisher)`
        a.read_publish = readout
        a.connect("COM5")  # connect to device
        # in listen everything happens until told to stop from elsewhere
        a.listen(stop_event)
        a.disconnect()

    Like the :class:`MessageHandler`, this class can be used as a context manager disconnecting at
    the end of the context.

    The (via RPC available) methods :meth:`get_parameters`, :meth:`set_parameters`, and
    :meth:`call_action` get/set parameters of the device or call an action of the device.
    You can also register device methods with :meth:`register_device_method`, such that this method
    is available via RPC as well.

    :param str name: Name to listen to and to publish values with.
    :param class device_class: Instrument class.
    :param int port: Port number to connect to.
    :param periodic_reading: Interval between periodic readouts in s.
    :param dict auto_connect: Kwargs to automatically connect to the device.
    :param class cls: See :code:`device_class`.

        .. deprecated:: 0.3
            Deprecated, use :code:`device_class` instead.

    :param \\**kwargs: Keywoard arguments for the general message handling.
    """

    device: Device

    def __init__(
        self,
        name: str,
        device_class: Optional[type[Device]] = None,
        periodic_reading: float = -1,
        auto_connect: Optional[dict] = None,
        context: Optional[zmq.Context] = None,
        cls: Optional[type[Device]] = None,
        **kwargs,
    ):
        context = context or zmq.Context.instance()
        super().__init__(name=name, context=context, **kwargs)
        if cls is not None:
            warn("Parameter `cls` is deprecated, use `device_class` instead.", FutureWarning)
            device_class = cls
        if device_class is None:
            # Keep this check as long as device_class is optional due to deprecated cls parameter
            raise ValueError("You have to specify a `device_class`!")
        self.device_class = device_class

        # Pipe for the periodic readout timer
        self.pipe: zmq.Socket = context.socket(zmq.PAIR)
        self.pipe.set_hwm(1)
        pipe_port = self.pipe.bind_to_random_port("inproc://listenerPipe", min_port=12345)
        self.pipeL: zmq.Socket = context.socket(zmq.PAIR)
        self.pipeL.set_hwm(1)
        self.pipeL.connect(f"inproc://listenerPipe:{pipe_port}")

        self.timer = RepeatingTimer(interval=periodic_reading, function=self.queue_readout)
        self.publisher = DataPublisher(full_name=name, log=self.root_logger)

        if auto_connect:
            self.connect(**auto_connect)
        self.log.info(f"Actor '{name}' initialized.")

    def register_rpc_methods(self) -> None:
        super().register_rpc_methods()
        self.register_rpc_method(self.get_parameters)
        self.register_rpc_method(self.set_parameters)
        self.register_rpc_method(self.call_action)
        self.register_rpc_method(self.start_polling)
        self.register_rpc_method(self.stop_polling)
        self.register_rpc_method(self.get_polling_interval)
        self.register_rpc_method(self.set_polling_interval)
        self.register_rpc_method(self.connect)
        self.register_rpc_method(self.disconnect)

    def register_device_method(self, method: Callable) -> None:
        """Make a device method available via RPC. The method name is prefixed with `device.`."""
        # TODO TBD how to call a device method?
        name = method.__name__
        self.register_rpc_method(method=method, name="device." + name)

    def __del__(self) -> None:
        self.disconnect()

    def __exit__(self, *args, **kwargs) -> None:
        super().__exit__(*args, **kwargs)
        self.disconnect()

    def set_full_name(self, full_name: str) -> None:
        super().set_full_name(full_name=full_name)
        self.publisher.full_name = full_name

    def _listen_setup(self) -> zmq.Poller:
        """Setup for listening."""
        poller = super()._listen_setup()
        poller.register(self.pipeL, zmq.POLLIN)
        return poller

    def _listen_loop_element(self, poller: zmq.Poller, waiting_time: Optional[int]
                             ) -> dict[zmq.Socket, int]:
        """Check the socks for incoming messages and handle them.

        :param waiting_time: Timeout of the poller in ms.
        """
        socks = super()._listen_loop_element(poller, waiting_time)
        if self.pipeL in socks:
            self.pipeL.recv()
            self.readout()
            del socks[self.pipeL]
        return socks

    def queue_readout(self) -> None:
        self.pipe.send(b"")

    def read_publish(self, device: Device, publisher: DataPublisher) -> None:
        """Read the device and publish the results.

        Defaults to doing nothing. Implement in a subclass.
        """
        self.log.warning("No 'read_publish' method defined, periodic readout does nothing.")

    def readout(self) -> None:
        """Do periodic readout of the instrument and publish the data.

        Defaults to calling :meth:`read_publish` with the device and publisher as arguments.
        """
        self.read_publish(device=self.device, publisher=self.publisher)

    def start_timer(self, interval: Optional[float] = None) -> None:
        """Start the readout timer.

        :param interval: Readout interval in s. If None, use the last value.
        """
        if interval is not None:
            self.timer.interval = interval
        if self.timer.interval < 0:
            return
        try:
            self.timer.start()
        except RuntimeError:
            self.timer = RepeatingTimer(interval=self.timer.interval, function=self.queue_readout)
            self.timer.start()

    def stop_timer(self) -> None:
        """Stop the readout timer."""
        try:
            self.timer.cancel()
        except AttributeError:
            pass

    def start_polling(self, polling_interval: Optional[float] = None) -> None:
        self.start_timer(interval=polling_interval)

    def stop_polling(self) -> None:
        self.stop_timer()

    @property
    def polling_interval(self) -> float:
        """Timeout interval of the readout timer in s."""
        return self.timer.interval

    @polling_interval.setter
    def polling_interval(self, value: float) -> None:
        self.timer.interval = value

    def get_polling_interval(self) -> float:
        return self.polling_interval

    def set_polling_interval(self, polling_interval: float) -> None:
        self.polling_interval = polling_interval

    def connect(self, *args, **kwargs) -> None:
        """Connect to the device with the given arguments and keyword arguments."""
        # TODO read auto_connect?
        self.log.info("Connecting")
        self.device = self.device_class(*args, **kwargs)
        self.start_timer()

    def disconnect(self) -> None:
        """Disconnect the device."""
        self.log.info("Disconnecting.")
        self.stop_timer()
        try:
            # Assumes a pymeasure instrument
            self.device.adapter.close()  # type: ignore
        except AttributeError:
            pass
        try:
            del self.device
        except AttributeError:
            pass

    def get_parameters(self, parameters: Union[list[str], tuple[str, ...]]) -> dict[str, Any]:
        """Get device properties from the list `properties`."""
        # `parameters` should be `Iterable[str]`, however, openrpc does not like that.
        data = {}
        for key in parameters:
            path = key.split(".")
            v = self.device
            for attr in path:
                v = getattr(v, attr)
            if callable(v):
                raise TypeError(f"Attribute '{key}' is a callable!")
            data[key] = v
        return data

    def set_parameters(self, parameters: dict[str, Any]) -> None:
        """Set device properties from a dictionary."""
        for key, value in parameters.items():
            path = key.split(".")
            obj = self.device
            for attr in path[:-1]:
                obj = getattr(obj, attr)
            setattr(obj, path[-1], value)

    def call_action(self, action: str, args: Optional[Sequence] = None,
                    kwargs: Optional[dict[str, Any]] = None) -> Any:
        """Call a device action with positional ``args`` and keyword arguments ``kwargs``."""
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        path = action.split(".")
        obj = self.device
        for attr in path[:-1]:
            obj = getattr(obj, attr)
        return getattr(obj, path[-1])(*args, **kwargs)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from typing import Any, Generic, Optional, Sequence, TypeVar, Union

from zmq import Context

from .actor import Actor


Device = TypeVar("Device")


class AccessDeniedError(BaseException):
    pass


class LockingActor(Actor, Generic[Device]):
    """An Actor which allows to lock the device or parts of it."""

    def __init__(
        self,
        name: str,
        device_class: Optional[type[Device]] = None,
        periodic_reading: float = -1,
        auto_connect: Optional[dict] = None,
        context: Optional[Context] = None,
        **kwargs,
    ):
        super().__init__(name, device_class, periodic_reading, auto_connect, context, **kwargs)
        self._locks: dict[Optional[str], bytes] = {}

    def register_rpc_methods(self) -> None:
        super().register_rpc_methods()
        self.register_rpc_method(self.lock)
        self.register_rpc_method(self.unlock)
        self.register_rpc_method(self.force_unlock)

    # RPC methods for locking
    def lock(self, resource: Optional[str] = None) -> bool:
        """Lock the controlled device or one of its resources and return the success state."""
        if self.check_access_rights(resource):
            self._locks[resource] = self.current_message.sender
            return True
        else:
            return False

    def unlock(self, resource: Optional[str] = None) -> None:
        """Unlock the controlled device or one of its resources.

        Only the locking Component may unlock.
        """
        if self.check_access_rights(resource):
            self._locks.pop(resource, None)

    def force_unlock(self, resource: Optional[str] = None) -> None:
        """Unlock the controlled device or one of its resources even if someone else locked it."""
        self._locks.pop(resource, None)

    # modified methods for device access
    def get_parameters(self, parameters: Union[list[str], tuple[str, ...]]) -> dict[str, Any]:
        # `parameters` should be `Iterable[str]`, however, openrpc does not like that.
        for parameter in parameters:
            self._check_access_rights_raising(parameter)
        return super().get_parameters(parameters=parameters)

    def set_parameters(self, parameters: dict[str, Any]) -> None:
        for parameter in parameters.keys():
            self._check_access_rights_raising(parameter)
        return super().set_parameters(parameters=parameters)

    def call_action(
        self, action: str, args: Optional[Sequence] = None, kwargs: Optional[dict[str, Any]] = None
    ) -> Any:
        self._check_access_rights_raising(action)
        return super().call_action(action=action, args=args, kwargs=kwargs)

    # helper methods
    def check_access_rights(self, resource: Optional[str]) -> bool:
        requester = self.current_message.sender
        if resource is None:
            elements = []
        else:
            elements = resource.split(".")
        for i in range(-1, len(elements)):
            local_owner = self._locks.get(".".join(elements[:i + 1])) if i >= 0 else None
            if local_owner is not None and requester != local_owner:
                return False
        return True

    def _check_access_rights_raising(self, resource: str) -> None:
        if self.check_access_rights(resource=resource) is False:
            raise AccessDeniedError(f"Resource '{resource}' is locked by someone else.")



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import json
import logging

import pytest

from pyleco.json_utils.rpc_generator import RPCGenerator
from pyleco.json_utils.rpc_server import RPCServer
from pyleco.json_utils.json_objects import (
    Request,
    ParamsRequest,
    ResultResponse,
    ErrorResponse,
    DataError,
    ResponseBatch,
)
from pyleco.json_utils.errors import (
    ServerError,
    InvalidRequest,
    INVALID_REQUEST,
    SERVER_ERROR,
    INTERNAL_ERROR,
)


@pytest.fixture
def rpc_generator() -> RPCGenerator:
    return RPCGenerator()


args = None
def side_effect_method(arg=None) -> int:
    global args
    args = (arg,)
    return 5


def fail() -> None:
    """Fail always.

    This method fails always.
    And has a description.
    """
    raise NotImplementedError


def simple() -> int:
    """A method without parameters."""
    return 7


def obligatory_parameter(arg1: float) -> float:
    """Needs an argument."""
    return arg1 * 2


@pytest.fixture
def rpc_server() -> RPCServer:
    rpc_server = RPCServer()
    rpc_server.method(name="sem")(side_effect_method)
    rpc_server.method()(side_effect_method)
    rpc_server.method()(fail)
    rpc_server.method()(simple)
    rpc_server.method()(obligatory_parameter)
    return rpc_server


def test_success(rpc_generator: RPCGenerator, rpc_server: RPCServer):
    request = rpc_generator.build_request_str(method="sem", arg=3)
    response = rpc_server.process_request(request)
    result = rpc_generator.get_result_from_response(response)  # type: ignore
    assert result == 5
    assert args == (3,)  # type: ignore


def test_multiple_requests_success(rpc_server: RPCServer, rpc_generator: RPCGenerator):
    request1 = ParamsRequest(id=1, method="sem", params=[3])
    request2 = Request(id=2, method="side_effect_method")
    message = json.dumps([request1.model_dump(), request2.model_dump()])
    result = rpc_server.process_request(message)
    result_obj = json.loads(result)  # type: ignore
    assert rpc_generator.get_result_from_response(result_obj[0]) == 5
    assert rpc_generator.get_result_from_response(result_obj[1]) == 5


def test_failing_method(rpc_generator: RPCGenerator, rpc_server: RPCServer):
    request = rpc_generator.build_request_str(method="fail")
    response = rpc_server.process_request(request)
    with pytest.raises(ServerError) as exc_info:
        rpc_generator.get_result_from_response(response)  # type: ignore
    error = exc_info.value.rpc_error
    assert error.code == SERVER_ERROR.code
    assert error.message == SERVER_ERROR.message


@pytest.mark.xfail(True, reason="Self written RPCServer cannot handle additional args")
def test_wrong_method_arguments_are_ignored(rpc_generator: RPCGenerator, rpc_server: RPCServer):
    request = rpc_generator.build_request_str(method="simple", arg=9)
    response = rpc_server.process_request(request)
    result = rpc_generator.get_result_from_response(response)  # type: ignore
    assert result == 7


def test_wrong_method_arguments_raise_error(rpc_generator: RPCGenerator, rpc_server: RPCServer):
    request = rpc_generator.build_request_str(method="simple", arg=9)
    response = rpc_server.process_request(request)
    response_obj = json.loads(response)  # type: ignore
    expected_response_obj = ErrorResponse(id=1, error=SERVER_ERROR)
    assert response_obj == expected_response_obj.model_dump()


def test_obligatory_parameter_missing(rpc_generator: RPCGenerator, rpc_server: RPCServer):
    request = rpc_generator.build_request_str(method="obligatory_parameter")
    response = rpc_server.process_request(request)
    with pytest.raises(ServerError) as exc_info:
        rpc_generator.get_result_from_response(response)  # type: ignore
    error = exc_info.value.rpc_error
    assert error.code == SERVER_ERROR.code
    assert error.message == SERVER_ERROR.message


def test_process_response(rpc_server: RPCServer, rpc_generator: RPCGenerator):
    """It should be a request, not a response!"""
    request = ResultResponse(id=7, result=9)
    request_string = request.model_dump_json()
    response = rpc_server.process_request(request_string)
    with pytest.raises(InvalidRequest) as exc_info:
        rpc_generator.get_result_from_response(response)  # type: ignore
    error = exc_info.value.rpc_error
    assert error.code == INVALID_REQUEST.code
    assert error.message == INVALID_REQUEST.message
    assert error.data == request.model_dump()  # type: ignore


class Test_discover_method:

    @pytest.fixture
    def discovered(self, rpc_server: RPCServer, rpc_generator: RPCGenerator) -> dict:
        request = rpc_generator.build_request_str("rpc.discover")
        response = rpc_server.process_request(request)
        return rpc_generator.get_result_from_response(response)  # type: ignore

    def test_info(self, discovered: dict):
        info: dict = discovered["info"]
        assert info.get("title") == "RPC Server"
        assert info.get("version") == "0.1.0"

    @pytest.fixture
    def methods(self, discovered: dict) -> list:
        return discovered.get("methods")  # type: ignore

    def test_side_effect_method(self, methods: list):
        m1: dict = methods[0]
        m2 = methods[1]
        assert m1["name"] == "sem"
        assert m1.get("description") is None
        assert m2["name"] == "side_effect_method"

    def test_fail(self, methods: list):
        m = methods[2]
        assert m["name"] == "fail"
        assert m["summary"] == "Fail always."
        assert m.get("description") == "This method fails always. And has a description."

    def test_simple_description(self, methods: list):
        m = methods[3]
        assert m["name"] == "simple"
        assert m["summary"] == "A method without parameters."
        assert m.get("description") is None

    def test_rpc_discover_not_listed(self, methods: list):
        for m in methods:
            if m.get("name") == "rpc.discover":
                raise AssertionError("rpc.discover is listed as a method!")


# tests regarding the local implementation of the RPC Server
def test_process_single_notification(rpc_server: RPCServer):
    result = rpc_server._process_single_request(
        {"jsonrpc": "2.0", "method": "simple"}
    )
    assert result is None


class Test_process_request:
    def test_log_exception(self, rpc_server: RPCServer, caplog: pytest.LogCaptureFixture):
        rpc_server.process_request(b"\xff")
        records = caplog.record_tuples
        assert records[-1] == (
            "pyleco.json_utils.rpc_server",
            logging.ERROR,
            "UnicodeDecodeError:",
        )

    def test_exception_response(self, rpc_server: RPCServer):
        result = rpc_server.process_request(b"\xff")
        assert result == ErrorResponse(id=None, error=INTERNAL_ERROR).model_dump_json()

    def test_invalid_request(self, rpc_server: RPCServer):
        result = rpc_server.process_request(b"7")
        assert (
            result
            == ErrorResponse(
                id=None, error=DataError.from_error(INVALID_REQUEST, 7)
            ).model_dump_json()
        )

    def test_batch_entry_notification(self, rpc_server: RPCServer):
        """A notification (request without id) shall not return anything."""
        requests = [
            {"jsonrpc": "2.0", "method": "simple"},
            {"jsonrpc": "2.0", "method": "simple", "id": 4},
        ]
        result = json.loads(rpc_server.process_request(json.dumps(requests)))  # type: ignore
        assert result == [{"jsonrpc": "2.0", "result": 7, "id": 4}]

    def test_batch_of_notifications(self, rpc_server: RPCServer):
        """A notification (request without id) shall not return anything."""
        requests = [
            {"jsonrpc": "2.0", "method": "simple"},
            {"jsonrpc": "2.0", "method": "simple"},
        ]
        result = rpc_server.process_request(json.dumps(requests))
        assert result is None

    def test_notification(self, rpc_server: RPCServer):
        """A notification (request without id) shall not return anything."""
        requests = [
            {"jsonrpc": "2.0", "method": "simple"},
        ]
        result = rpc_server.process_request(json.dumps(requests))
        assert result is None


class Test_process_request_object:
    def test_invalid_request(self, rpc_server: RPCServer):
        result = rpc_server.process_request_object(7)
        assert (
            result
            == ErrorResponse(
                id=None, error=DataError.from_error(INVALID_REQUEST, 7)
            )
        )

    def test_batch_entry_notification(self, rpc_server: RPCServer):
        """A notification (request without id) shall not return anything."""
        requests = [
            {"jsonrpc": "2.0", "method": "simple"},
            {"jsonrpc": "2.0", "method": "simple", "id": 4},
        ]
        result = rpc_server.process_request_object(requests)
        assert result == ResponseBatch([ResultResponse(4, 7)])

    def test_batch_of_notifications(self, rpc_server: RPCServer):
        """A notification (request without id) shall not return anything."""
        requests = [
            {"jsonrpc": "2.0", "method": "simple"},
            {"jsonrpc": "2.0", "method": "simple"},
        ]
        result = rpc_server.process_request_object(requests)
        assert result is None

    def test_notification(self, rpc_server: RPCServer):
        """A notification (request without id) shall not return anything."""
        requests = [
            {"jsonrpc": "2.0", "method": "simple"},
        ]
        result = rpc_server.process_request_object(requests)
        assert result is None



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.test import FakeCommunicator, FakeContext
from pyleco.core.message import Message, MessageTypes

try:
    from pyleco.utils.qt_listener import QtListener, QtPipeHandler, ListenerSignals
except ModuleNotFoundError:
    pytest.skip(reason="qtpy not installed.", allow_module_level=True)

cid = b"conversation_id;"


@pytest.fixture
def signal():
    class FakeSignal:
        def emit(self, message: Message):
            self._content = message
    return FakeSignal()


@pytest.fixture
def qt_listener(signal) -> QtListener:
    qt_listener = QtListener(name="test")  # type: ignore
    qt_listener.communicator = FakeCommunicator(name="N.Pipe")  # type: ignore
    qt_listener.signals.message = signal
    return qt_listener


@pytest.fixture
def qt_handler(signal) -> QtPipeHandler:
    handler = QtPipeHandler(name="handler",
                            context=FakeContext(),  # type: ignore
                            signals=ListenerSignals())
    handler.signals.message = signal
    return handler


class Test_handle_message:
    def test_handle_valid_jsonrpc(self, qt_handler: QtPipeHandler):
        msg = Message("N.Pipe", "sender",
                      data={"jsonrpc": "2.0", "method": "abc", "id": 6},
                      message_type=MessageTypes.JSON,
                      conversation_id=cid,
                      )
        qt_handler.handle_message(msg)
        assert qt_handler.signals.message._content == msg  # type: ignore

    def test_empty_message(self, qt_handler: QtPipeHandler):
        msg = Message("N.Pipe", "sender")
        qt_handler.handle_message(msg)
        assert qt_handler.signals.message._content == msg  # type: ignore

    def test_local_method(self, qt_handler: QtPipeHandler):
        msg = Message("handler", "sender",
                      conversation_id=cid, message_type=MessageTypes.JSON,
                      data={"jsonrpc": "2.0", "method": "pong", "id": 3})
        qt_handler.handle_message(msg)
        assert Message.from_frames(*qt_handler.socket._s[0]) == Message(  # type: ignore
            "sender", "handler", conversation_id=cid, message_type=MessageTypes.JSON,
            data={"jsonrpc": "2.0", "result": None, "id": 3}
        )




#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

"""
Names based on the classes of jsonrpc2-objects.

As jsonrpc2-objects uses pydantic models, these objects offer the `dump_model` and `dump_model_json`
methods.
"""

from __future__ import annotations
from dataclasses import asdict, dataclass
import json
from typing import Any, List, Optional, TypeVar, Union

ErrorType = Union["DataError", "Error"]
NotificationType = Union["Notification", "ParamsNotification"]
RequestType = Union["ParamsRequest", "Request"]
ResponseType = Union["ErrorResponse", "ResultResponse"]


@dataclass
class JsonObject:
    def model_dump(self) -> dict[str, Any]:
        """Create a dictionary of the attributes."""
        return asdict(self)

    def model_dump_json(self) -> str:
        """Create a json representation."""
        return json.dumps(self.model_dump(), separators=(",", ":"))


@dataclass
class Request(JsonObject):
    """Request the result of a remote call."""

    id: Union[int, str]
    method: str
    jsonrpc: str = "2.0"


@dataclass
class ParamsRequest(JsonObject):
    """Request the result of a remote call with parameters."""

    id: Union[int, str]
    method: str
    params: Union[list, dict]
    jsonrpc: str = "2.0"


@dataclass
class Notification(JsonObject):
    """Do a remote call without requesting a response."""

    method: str
    jsonrpc: str = "2.0"


@dataclass
class ParamsNotification(JsonObject):
    """Do a remote call with parameters without requesting a response."""

    method: str
    params: Union[list, dict]
    jsonrpc: str = "2.0"


@dataclass
class ResultResponse(JsonObject):
    """A response containing a result."""

    id: Union[int, str]
    result: Any
    jsonrpc: str = "2.0"


@dataclass
class Error(JsonObject):
    """An error to be sent via an :class:`ErrorResponse`."""

    code: int
    message: str


@dataclass
class DataError(JsonObject):
    """An error with data, to be sent via an :class:`ErrorResponse`."""

    code: int
    message: str
    data: Any

    @classmethod
    def from_error(cls, error: Error, data: Any) -> DataError:
        return cls(code=error.code, message=error.message, data=data)


@dataclass
class ErrorResponse(JsonObject):
    """A response containing an error."""

    id: Optional[Union[int, str]]
    error: ErrorType
    jsonrpc: str = "2.0"

    def model_dump(self) -> dict[str, Any]:
        pre_dict = asdict(self)
        pre_dict["error"] = asdict(self.error)
        return pre_dict

"""
Batch Handling.

Not included in jsonrpc2-objects, but defined by JSONRPC 2.0
"""
BatchType = TypeVar("BatchType", RequestType, ResponseType)


class BatchObject(List[BatchType]):
    """A batch of JSONRPC message objects.

    It works like a list of appropriate message objects and offers the possibility to dump
    this batch object to a plain python object or to JSON.
    """
    # Parent class is typing.List, as Python<3.9 does not like list[BatchType]
    # Not defined by jsonrpc2-objects

    def model_dump(self) -> list[dict[str, Any]]:
        return [obj.model_dump() for obj in self]

    def model_dump_json(self) -> str:
        return json.dumps(self.model_dump(), separators=(",", ":"))


RequestBatch = BatchObject[RequestType]
ResponseBatch = BatchObject[ResponseType]



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import logging
from sys import version_info
import time

from unittest.mock import MagicMock

import pytest

from pyleco.test import FakeSocket, FakePoller, assert_response_is_result, handle_request_message
from pyleco.actors.actor import Actor
from pyleco.core.leco_protocols import PollingActorProtocol, ExtendedComponentProtocol, Protocol


class FantasyChannel:
    def __init__(self) -> None:
        self._prop = -1

    @property
    def channel_property(self):
        return self._prop

    @channel_property.setter
    def channel_property(self, value):
        self._prop = value

    def channel_method(self, value):
        return 2 * value


class FantasyInstrument:

    def __init__(self, adapter, name="FantasyInstrument", *args, **kwargs):
        self.name = name
        self.adapter = adapter
        super().__init__()
        self._prop = 5
        self._prop2 = 7
        self.channel = FantasyChannel()
        self.channel.trace = FantasyChannel()  # type: ignore

    @property
    def prop(self):
        return self._prop

    @prop.setter
    def prop(self, value):
        self._prop = value

    @property
    def prop2(self):
        return self._prop2

    @prop2.setter
    def prop2(self, value):
        self._prop2 = value

    def silent_method(self, value):
        self._method_value = value

    def returning_method(self, value):
        return value ** 2

    @property
    def long(self):
        time.sleep(0.5)
        return 7

    def connect(self, *args):
        pass

    def disconnect(self, *args):
        pass


class FakeActor(Actor):

    def queue_readout(self):
        logging.getLogger().info(f"queue: {time.perf_counter()}")
        super().queue_readout()

    def heartbeat(self):
        logging.getLogger().info("beating")
        super().heartbeat()


class ExtendedActorProtocol(ExtendedComponentProtocol, PollingActorProtocol, Protocol):
    pass


@pytest.fixture()
def actor() -> FakeActor:
    actor = FakeActor(
        "test",
        FantasyInstrument,
        auto_connect={"adapter": MagicMock()},
        port=1234,
        protocol="inproc",
    )
    actor.socket = FakeSocket(5)  # type: ignore
    actor.next_beat = float("inf")
    return actor


class TestProtocolImplemented:
    protocol_methods = [m for m in dir(ExtendedActorProtocol) if not m.startswith("_")]

    def static_test_methods_are_present(self):
        def testing(component: ExtendedActorProtocol):
            pass
        testing(FakeActor(name="test", device_class=FantasyInstrument))

    @pytest.fixture
    def component_methods(self, actor: Actor):
        response = actor.rpc.process_request(
            '{"id": 1, "method": "rpc.discover", "jsonrpc": "2.0"}')
        result = actor.rpc_generator.get_result_from_response(response)  # type: ignore
        return result.get('methods')

    @pytest.mark.parametrize("method", protocol_methods)
    def test_method_is_available(self, component_methods, method):
        for m in component_methods:
            if m.get('name') == method:
                return
        raise AssertionError(f"Method {method} is not available.")


@pytest.mark.skipif(version_info.minor < 9,
                    reason="It is deprecated, because it does not work for python<3.9.")
def test_deprecated_cls_argument():
    with pytest.warns(FutureWarning, match="`cls` is deprecated"):
        actor = FakeActor("test", cls=FantasyInstrument, auto_connect={'adapter': MagicMock()},
                          port=1234,
                          protocol="inproc")
        assert actor.device_class == FantasyInstrument


def test_device_class_or_cls_is_necessary():
    with pytest.raises(ValueError, match="`device_class`"):
        FakeActor("test", protocol="inproc")


def test_get_parameters(actor: Actor):
    handle_request_message(actor, "get_parameters", ["prop"])
    result = assert_response_is_result(actor)
    assert result == {"prop": 5}


def test_get_channel_parameters(actor: Actor):
    handle_request_message(actor, "get_parameters", ["channel.channel_property"])
    result = assert_response_is_result(actor)
    assert result == {"channel.channel_property": -1}


def test_get_nested_channel_parameters(actor: Actor):
    handle_request_message(actor, "get_parameters", ["channel.trace.channel_property"])
    result = assert_response_is_result(actor)
    assert result == {"channel.trace.channel_property": -1}


def test_set_parameters(actor: Actor):
    handle_request_message(actor, "set_parameters", {"prop2": 10})
    assert_response_is_result(actor)
    assert actor.device.prop2 == 10


def test_set_channel_parameters(actor: Actor):
    handle_request_message(actor, "set_parameters", {"channel.channel_property": 10})
    assert_response_is_result(actor)
    assert actor.device.channel.channel_property == 10

def test_set_nested_channel_parameters(actor: Actor):
    handle_request_message(actor, "set_parameters", {"channel.trace.channel_property": 10})
    assert_response_is_result(actor)
    assert actor.device.channel.trace.channel_property == 10  # type: ignore


def test_call_silent_method(actor: Actor):
    handle_request_message(actor, "call_action", action="silent_method", kwargs=dict(value=7))
    assert_response_is_result(actor)
    assert actor.device._method_value == 7


def test_returning_method(actor: Actor):
    handle_request_message(actor, "call_action", action="returning_method", kwargs=dict(value=2))
    result = assert_response_is_result(actor)
    assert result == 4


def test_channel_method(actor: Actor):
    handle_request_message(actor, "call_action", action="channel.channel_method", args=(7,))
    result = assert_response_is_result(actor)
    assert result == 14


def test_nested_channel_method(actor: Actor):
    handle_request_message(actor, "call_action", action="channel.trace.channel_method", args=(7,))
    result = assert_response_is_result(actor)
    assert result == 14


def test_register_device_method(actor: Actor):
    actor.register_device_method(actor.device.returning_method)
    handle_request_message(actor, "device.returning_method", 5)
    result = assert_response_is_result(actor)
    assert result == 25


class Test_disconnect:
    @pytest.fixture
    def disconnected_actor(self):
        actor = FakeActor("name", device_class=FantasyInstrument,
                          auto_connect={"adapter": MagicMock()})
        actor._device = actor.device  # type: ignore
        actor.device.adapter.close = MagicMock()
        actor.disconnect()
        return actor

    def test_device_deleted(self, disconnected_actor: Actor):
        assert not hasattr(disconnected_actor, "device")

    def test_timer_canceled(self, disconnected_actor: Actor):
        assert disconnected_actor.timer.finished.is_set() is True

    def test_device_closed(self, disconnected_actor: Actor):
        disconnected_actor._device.adapter.close.assert_called_once()  # type: ignore


def test_exit_calls_disconnect():
    with FakeActor("name", device_class=FantasyInstrument) as actor:
        actor.disconnect = MagicMock()
    actor.disconnect.assert_called_once()


class Test_listen_loop_element:
    @pytest.fixture
    def looped_actor(self, actor: Actor):
        """Check a loop with a value in the pipe"""
        poller = FakePoller()
        poller.register(actor.pipeL)
        actor.queue_readout()  # enqueue a readout
        actor.readout = MagicMock()  # type: ignore
        # act
        socks = actor._listen_loop_element(poller=poller,  # type: ignore
                                           waiting_time=None)
        actor._socks = socks  # type: ignore  # for assertions
        return actor

    def test_socks_empty(self, looped_actor: Actor):
        assert looped_actor._socks == {}  # type: ignore

    def test_readout_called(self, looped_actor: Actor):
        looped_actor.readout.assert_called_once()  # type: ignore

    def test_no_readout_queued(self, actor: Actor):

        poller = FakePoller()
        poller.register(actor.pipeL)
        actor.readout = MagicMock()  # type: ignore
        # act
        actor._listen_loop_element(poller=poller,  # type: ignore
                                   waiting_time=0)
        actor.readout.assert_not_called()


def test_timer_enqueues_heartbeat(actor: Actor):
    actor.start_timer(0.0000001)  # s
    assert actor.pipeL.poll(timeout=50) == 1  # ms


def test_restart_stopped_timer(actor: Actor):
    """Starting a stopped timer is impossible, ensure, that it works as expected (new timer)."""
    actor.start_timer(10)  # s
    actor.stop_timer()
    actor.start_timer(0.0000001)  # s
    assert actor.pipeL.poll(timeout=50) == 1  # ms



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pickle
import pytest

from pyleco.utils.data_publisher import DataPublisher, DataMessage
from pyleco.test import FakeContext


@pytest.fixture
def publisher():
    publisher = DataPublisher(full_name="sender", context=FakeContext())  # type: ignore
    return publisher


def test_socket_type(publisher: DataPublisher):
    assert publisher.socket.socket_type == 1


def test_connection():
    publisher = DataPublisher(full_name="", host="localhost", port=12345,
                              context=FakeContext())  # type: ignore
    assert publisher.socket.addr == "tcp://localhost:12345"


def test_context_manager_closes_connection():
    with DataPublisher("", context=FakeContext()) as p:  # type: ignore
        pass
    assert p.socket.closed is True


def test_call_publisher_sends(publisher: DataPublisher):
    publisher(b"data")
    # assert
    message = DataMessage.from_frames(*publisher.socket._s.pop())  # type: ignore
    assert message.topic == publisher.full_name.encode()
    assert message.payload[0] == b"data"


def test_send_data(publisher: DataPublisher):
    publisher.send_data(
        data=b"data", topic=b"topic", conversation_id=b"cid", additional_payload=[b"1"]
    )
    assert publisher.socket._s == [[b"topic", b"cid\x00", b"data", b"1"]]


def test_send_message(publisher: DataPublisher):
    message = DataMessage.from_frames(b"topic", b"header", b"data")
    publisher.send_message(message=message)
    assert publisher.socket._s == [message.to_frames()]


def test_send_legacy(publisher: DataPublisher):
    value = 5.67
    publisher.send_legacy({'key': value})
    message = DataMessage.from_frames(*publisher.socket._s[0])  # type: ignore
    assert message.topic == b"key"
    assert message.payload[0] == pickle.dumps(value)
    assert message.message_type == 234


def test_set_full_name(publisher: DataPublisher):
    new_full_name = "new full name"
    publisher.set_full_name(new_full_name)
    assert publisher.full_name == new_full_name



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
import logging
from typing import Optional, Union

from .director import Director


log = logging.getLogger(__name__)
log.addHandler(logging.NullHandler())


class StarterDirector(Director):
    """Director for the Starter.

    :param actor: Name of the actor to direct.
    """

    def __init__(self, actor: str = "starter", **kwargs) -> None:
        super().__init__(actor=actor, **kwargs)

    def start_tasks(self, names: Union[list[str], str], actor: Optional[Union[bytes, str]] = None
                    ) -> None:
        """Start the task or tasks.

        :param names: Single task name or list of task names to start.
        :param name: Name of the starter to communicate with.
        """
        if isinstance(names, str):
            names = [names]
        self.ask_rpc(method="start_tasks", names=names, actor=actor)

    def restart_tasks(self, names: Union[list[str], str], actor: Optional[Union[bytes, str]] = None
                      ) -> None:
        """Restart the task or tasks.

        :param names: Single task name or list of task names to restart.
        :param name: Name of the starter to communicate with.
        """
        if isinstance(names, str):
            names = [names]
        self.ask_rpc(method="restart_tasks", names=names, actor=actor)

    def stop_tasks(self, names: Union[list[str], str], actor: Optional[Union[bytes, str]] = None
                   ) -> None:
        """Stop the task or tasks.

        :param names: Single task name or list of task names to stop.
        :param name: Name of the starter to communicate with.
        """
        if isinstance(names, str):
            names = [names]
        self.ask_rpc(method="stop_tasks", names=names, actor=actor)

    def install_tasks(self, names: Union[list[str], str], actor: Optional[Union[bytes, str]] = None
                      ) -> None:
        """Install the tasks.

        :param names: Single task name or list of task names to install.
        :param name: Name of the starter to communicate with.
        """
        if isinstance(names, str):
            names = [names]
        self.ask_rpc(method="install_tasks", names=names, actor=actor)

    def status_tasks(self, names: Optional[Union[list[str], str]] = None,
                     actor: Optional[Union[bytes, str]] = None) -> dict[str, int]:
        """Query the status of these tasks and all running ones.

        :param names: List of task names to ask for.
        :param name: Name of the starter to communicate with.
        """
        if isinstance(names, str):
            names = [names]
        return self.ask_rpc(method="status_tasks", names=names, actor=actor)

    def list_tasks(self, actor: Optional[Union[bytes, str]] = None) -> list[dict[str, str]]:
        """List all available tasks with name and tooltip."""
        return self.ask_rpc(method="list_tasks", actor=actor)



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from typing import Union

import pytest

from pyleco.json_utils.json_objects import ErrorResponse
from pyleco.json_utils.errors import (JSONRPCError, NODE_UNKNOWN, NOT_SIGNED_IN, DUPLICATE_NAME,
                                      RECEIVER_UNKNOWN)

from pyleco.json_utils.rpc_generator import RPCGenerator, INVALID_SERVER_RESPONSE


@pytest.fixture
def generator() -> RPCGenerator:
    return RPCGenerator()


@pytest.mark.parametrize("method, args, kwargs, result", (
        ("meth", (), {}, '{"id":1,"method":"meth","jsonrpc":"2.0"}'),
        ("with args", (5, ), {},
         '{"id":1,"method":"with args","params":[5],"jsonrpc":"2.0"}'),
        ("with kwargs", (), {'kwarg': 7},
         '{"id":1,"method":"with kwargs","params":{"kwarg":7},"jsonrpc":"2.0"}'),
))
def test_build_request_str(generator: RPCGenerator, method: str, args: tuple, kwargs: dict,
                           result: str) -> None:
    assert generator.build_request_str(method, *args, **kwargs) == result


def test_build_request_str_raises_error(generator: RPCGenerator):
    with pytest.raises(ValueError, match="same time"):
        generator.build_request_str("some_method", "argument", keyword="whatever")


@pytest.mark.parametrize("response, result", (
        ('{"id": 5, "result": 7.9, "jsonrpc": "2.0"}', 7.9),
        (b'{"id": 5, "result": 7.9, "jsonrpc": "2.0"}', 7.9),  # bytes version
        ('{"id": 7, "result": null, "jsonrpc": "2.0"}', None),
        ('{"id": 8, "result": [5, 8.9], "jsonrpc": "2.0"}', [5, 8.9]),
        ('{"id": 9, "result": "whatever", "jsonrpc": "2.0"}', "whatever"),
))
def test_get_result_from_response(generator: RPCGenerator, response: Union[bytes, str], result):
    assert generator.get_result_from_response(response) == result


@pytest.mark.parametrize("error", (NOT_SIGNED_IN, NODE_UNKNOWN, DUPLICATE_NAME, RECEIVER_UNKNOWN))
def test_get_result_from_response_raises_suitable_error(generator: RPCGenerator, error):
    with pytest.raises(JSONRPCError) as exc:
        generator.get_result_from_response(ErrorResponse(id=7, error=error).model_dump_json())
    assert exc.value.rpc_error == error


def test_invalid_response_raises_correct_error(generator: RPCGenerator):
    with pytest.raises(JSONRPCError) as exc:
        generator.get_result_from_response(b"\x00")
    error = exc.value.rpc_error
    assert error.code == INVALID_SERVER_RESPONSE.code
    assert error.message == INVALID_SERVER_RESPONSE.message



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.test import FakeContext
from pyleco.utils.zmq_log_handler import ZmqLogHandler, DataMessage


@pytest.fixture
def handler() -> ZmqLogHandler:
    return ZmqLogHandler(context=FakeContext(), full_name="fullname", port=12345)  # type: ignore


def test_init_(handler: ZmqLogHandler):
    assert handler.full_name == "fullname"


def test_init_address(handler: ZmqLogHandler):
    assert handler.queue.socket.addr == "tcp://localhost:12345"  # type: ignore


def test_enqueue(handler: ZmqLogHandler):
    handler.enqueue("whatever")
    message = DataMessage.from_frames(*handler.queue.socket._s.pop())  # type: ignore
    assert message.topic == b"fullname"
    assert message.payload == [b'whatever']



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import json
import pickle
from unittest.mock import MagicMock

import pytest

from pyleco.core.data_message import DataMessage
from pyleco.test import FakeContext, FakeSocket, handle_request_message, assert_response_is_result
from pyleco.utils.events import SimpleEvent
from pyleco.utils.extended_message_handler import ExtendedMessageHandler


@pytest.fixture
def handler():
    handler = ExtendedMessageHandler(name="handler",
                                     context=FakeContext())  # type: ignore
    handler.namespace = "N1"
    handler.stop_event = SimpleEvent()
    handler.subscriber = FakeSocket(2)  # type: ignore
    handler.handle_subscription_message = MagicMock()  # it is not defined
    return handler


def test_read_subscription_message_calls_handle(handler: ExtendedMessageHandler):
    message = DataMessage("", data="[]")
    handler.subscriber._r = [message.to_frames()]  # type: ignore
    handler.read_subscription_message()
    # assert
    handler.handle_subscription_message.assert_called_once_with(message)  # type: ignore


def test_handle_subscription_message_raises_not_implemented():
    handler = ExtendedMessageHandler(name="handler", context=FakeContext())  # type: ignore
    with pytest.raises(NotImplementedError):
        handler.handle_subscription_message(DataMessage(b"topic"))


def test_read_subscription_message_calls_handle_legacy(handler: ExtendedMessageHandler):
    message = DataMessage("", data="[]", message_type=234)
    handler.handle_full_legacy_subscription_message = MagicMock()  # type: ignore[method-assign]
    handler.subscriber._r = [message.to_frames()]  # type: ignore
    handler.read_subscription_message()
    # assert
    handler.handle_full_legacy_subscription_message.assert_called_once_with(message)  # type: ignore


def test_subscribe_single(handler: ExtendedMessageHandler):
    handler.subscribe_single(b"topic")
    assert handler.subscriber._subscriptions == [b"topic"]  # type: ignore
    assert handler._subscriptions == [b"topic"]


def test_subscribe_single_again(handler: ExtendedMessageHandler, caplog: pytest.LogCaptureFixture):
    # arrange
    handler.subscribe_single(b"topic")
    caplog.set_level(10)
    # act
    handler.subscribe_single(b"topic")
    assert caplog.messages[-1] == f"Already subscribed to {b'topic'!r}."


@pytest.mark.parametrize("topics, result", (
        ("topic", [b"topic"]),  # single string
        (["topic1", "topic2"], [b"topic1", b"topic2"]),  # list of strings
        (("topic1", "topic2"), [b"topic1", b"topic2"]),  # tuple of strings
))
def test_subscribe(handler: ExtendedMessageHandler, topics, result):
    handle_request_message(handler, "subscribe", topics)
    assert_response_is_result(handler)
    assert handler._subscriptions == result


def test_unsubscribe_single(handler: ExtendedMessageHandler):
    handler._subscriptions = [b"topic"]
    handler.subscriber._subscriptions = [b"topic"]  # type: ignore
    handler.unsubscribe_single(b"topic")
    assert handler._subscriptions == []
    assert handler.subscriber._subscriptions == []  # type: ignore


@pytest.mark.parametrize("topics, result", (
        ("topic", [b"topic"]),  # single string
        (["topic1", "topic2"], [b"topic1", b"topic2"]),  # list of strings
        (("topic1", "topic2"), [b"topic1", b"topic2"]),  # tuple of strings
))
def test_unsubscribe(handler: ExtendedMessageHandler, topics, result):
    handler._subscriptions = result
    # act
    handle_request_message(handler, "unsubscribe", topics)
    assert_response_is_result(handler)
    assert handler._subscriptions == []


def test_unsubscribe_all(handler: ExtendedMessageHandler):
    handler._subscriptions = [b"topic1", b"topic2"]
    # act
    handle_request_message(handler, "unsubscribe_all")
    # assert
    assert_response_is_result(handler)
    assert handler._subscriptions == []


class Test_handle_full_legacy_subscription_message:
    @pytest.fixture
    def handler_hfl(self, handler: ExtendedMessageHandler) -> ExtendedMessageHandler:
        handler.handle_subscription_data = MagicMock()  # type: ignore[method-assign]
        return handler

    def test_handle_pickled_message(self, handler_hfl: ExtendedMessageHandler):
        data = ["some", "data", 5]
        handler_hfl.handle_full_legacy_subscription_message(
            DataMessage("topic", data=pickle.dumps(data), message_type=234)
        )
        handler_hfl.handle_subscription_data.assert_called_once_with({"topic": data})  # type: ignore

    def test_handle_json_message(self, handler_hfl: ExtendedMessageHandler):
        data = ["some", "data", 5]
        handler_hfl.handle_full_legacy_subscription_message(
            DataMessage("topic", data=json.dumps(data), message_type=235)
        )
        handler_hfl.handle_subscription_data.assert_called_once_with({"topic": data})  # type: ignore

    def test_handle_unknown_message_type(self, handler_hfl: ExtendedMessageHandler):
        with pytest.raises(ValueError):
            handler_hfl.handle_full_legacy_subscription_message(
                DataMessage("topic", data="", message_type=210)
            )

    def test_handle_subscription_data(self, handler: ExtendedMessageHandler):
        with pytest.raises(NotImplementedError):
            handler.handle_subscription_data({})



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

import pytest

from pyleco.test import FakeDirector
from pyleco.coordinators.coordinator import Coordinator
from pyleco.directors.coordinator_director import CoordinatorDirector


class FakeCoordinatorDirector(FakeDirector, CoordinatorDirector):
    """Replaces the ask_rpc method."""


@pytest.fixture
def coordinator_director():
    data_logger_director = FakeCoordinatorDirector(remote_class=Coordinator)
    return data_logger_director


@pytest.mark.parametrize("method", ("get_local_components", "get_global_components", "get_nodes",
                                    ))
def test_method_call_existing_remote_methods(coordinator_director: FakeCoordinatorDirector, method):
    """Test that the remote method exists."""
    coordinator_director.return_value = None
    getattr(coordinator_director, method)()
    # asserts that no error is raised in the "ask_rpc" method



#
# This file is part of the PyLECO package.
#
# Copyright (c) 2023-2025 PyLECO Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#

from __future__ import annotations
from functools import wraps
from json import JSONDecodeError
import logging
import time
from typing import Any, Callable, Optional, Union, TypeVar

import zmq

from ..core import COORDINATOR_PORT
from ..core.leco_protocols import ExtendedComponentProtocol
from ..core.message import Message, MessageTypes
from ..core.serialization import JsonContentTypes, get_json_content_type
from ..json_utils.errors import JSONRPCError
from ..json_utils.rpc_generator import RPCGenerator
from ..json_utils.rpc_server import RPCServer
from .base_communicator import BaseCommunicator
from .log_levels import PythonLogLevels
from .zmq_log_handler import ZmqLogHandler
from .events import Event, SimpleEvent


# Parameters
heartbeat_interval = 10  # s


ReturnValue = TypeVar("ReturnValue")


class MessageHandler(BaseCommunicator, ExtendedComponentProtocol):
    """Maintain connection to the Coordinator and listen to incoming messages.

    This class is intended to run in a thread, maintain the connection to the coordinator
    with heartbeats and timely responses. If a message arrives which is not connected to the control
    protocol itself (e.g. ping message), another method is called.
    You may subclass this class in order to handle these messages as desired.

    You may use it as a context manager.

    :param str name: Name to listen to and to publish values with.
    :param str host: Host name (or IP address) of the Coordinator to connect to.
    :param int port: Port number of the Coordinator to connect to.
    :param protocol: Connection protocol.
    :param log: Logger instance whose logs should be published. Defaults to `getLogger("__main__")`.
    """

    name: str

    current_message: Message
    additional_response_payload: Optional[list[bytes]] = None

    def __init__(
        self,
        name: str,
        host: str = "localhost",
        port: int = COORDINATOR_PORT,
        protocol: str = "tcp",
        log: Optional[logging.Logger] = None,
        context: Optional[zmq.Context] = None,
        **kwargs,
    ):
        self.name = name
        self._namespace: Union[str, None] = None
        self._full_name: str = name
        self.rpc = RPCServer(title=name)
        self.rpc_generator = RPCGenerator()
        self.register_rpc_methods()

        self.setup_logging(log=log)
        self.setup_socket(
            host=host, port=port, protocol=protocol, context=context or zmq.Context.instance()
        )

        super().__init__(**kwargs)
        self.setup_message_buffer()

    @property
    def namespace(self) -> Union[str, None]:
        return self._namespace

    @namespace.setter
    def namespace(self, value: Union[str, None]) -> None:
        self._namespace = value
        full_name = self.name if value is None else ".".join((value, self.name))
        self.set_full_name(full_name=full_name)

    def set_full_name(self, full_name: str) -> None:
        self._full_name = full_name
        self.rpc.title = full_name
        self.log_handler.full_name = full_name

    @property
    def full_name(self) -> str:
        return self._full_name

    def setup_logging(self, log):
        if log is None:
            log = logging.getLogger("__main__")
        # Add the ZmqLogHandler to the root logger, unless it has already a Handler.
        first_pub_handler = True  # we expect to be the first ZmqLogHandler
        for h in log.handlers:
            if isinstance(h, ZmqLogHandler):
                first_pub_handler = False
                self.log_handler = h
                break
        if first_pub_handler:
            self.log_handler = ZmqLogHandler()
            log.addHandler(self.log_handler)
        self.root_logger = log
        self.log = self.root_logger.getChild("MessageHandler")  # for cooperation

    def setup_socket(self, host: str, port: int, protocol: str, context: zmq.Context) -> None:
        self.socket: zmq.Socket = context.socket(zmq.DEALER)
        self.log.info(f"MessageHandler connecting to {host}:{port}")
        self.socket.connect(f"{protocol}://{host}:{port}")

    def register_rpc_method(self, method: Callable[..., Any], **kwargs) -> None:
        """Register a method to be available via rpc calls."""
        self.rpc.method(**kwargs)(method)

    def _handle_binary_return_value(
        self, return_value: tuple[ReturnValue, list[bytes]]
    ) -> ReturnValue:
        self.additional_response_payload = return_value[1]
        return return_value[0]

    @staticmethod
    def _pass_through(return_value: ReturnValue) -> ReturnValue:
        return return_value

    def _generate_binary_capable_method(
        self,
        method: Callable[..., Union[ReturnValue, tuple[ReturnValue, list[bytes]]]],
        accept_binary_input: bool = False,
        return_binary_output: bool = False,
    ) -> Callable[..., ReturnValue]:
        returner = self._handle_binary_return_value if return_binary_output else self._pass_through
        if accept_binary_input is True:

            @wraps(method)
            def modified_method(*args, **kwargs) -> ReturnValue:  # type: ignore
                if args:
                    args_l = list(args)
                    if args_l[-1] is None:
                        args_l[-1] = self.current_message.payload[1:]
                    else:
                        args_l.append(self.current_message.payload[1:])
                    args = args_l  # type: ignore[assignment]
                else:
                    kwargs["additional_payload"] = self.current_message.payload[1:]
                return_value = method(
                    *args, **kwargs
                )
                return returner(return_value=return_value)  # type: ignore
        else:

            @wraps(method)
            def modified_method(*args, **kwargs) -> ReturnValue:
                return_value = method(*args, **kwargs)
                return returner(return_value=return_value)  # type: ignore

        doc_addition = (
            f"(binary{' input' * accept_binary_input}{' output' * return_binary_output} method)"
        )
        try:
            modified_method.__doc__ += "\n" + doc_addition  # type: ignore[operator]
        except TypeError:
            modified_method.__doc__ = doc_addition
        return modified_method  # type: ignore

    def register_binary_rpc_method(
        self,
        method: Callable[..., Union[Any, tuple[Any, list[bytes]]]],
        accept_binary_input: bool = False,
        return_binary_output: bool = False,
        **kwargs,
    ) -> None:
        """Register a method which accepts binary input and/or returns binary values.

        :param accept_binary_input: the method must accept the additional payload as an
            `additional_payload=None` parameter (default value must be present as `None`!).
        :param return_binary_output: the method must return a tuple of a JSON-able python object
            (e.g. `None`) and of a list of bytes objects, to be sent as additional payload.
        """
        modified_method = self._generate_binary_capable_method(
            method=method,
            accept_binary_input=accept_binary_input,
            return_binary_output=return_binary_output,
        )
        self.register_rpc_method(modified_method, **kwargs)

    def register_rpc_methods(self) -> None:
        """Register methods for RPC."""
        self.register_rpc_method(self.shut_down)
        self.register_rpc_method(self.set_log_level)
        self.register_rpc_method(self.pong)

    # Base communication
    def send(
        self,
        receiver: Union[bytes, str],
        conversation_id: Optional[bytes] = None,
        data: Optional[Any] = None,
        **kwargs,
    ) -> None:
        """Send a message to a receiver with serializable `data`."""
        try:
            super().send(receiver=receiver, conversation_id=conversation_id, data=data, **kwargs)
        except Exception as exc:
            self.log.exception(f"Composing message with data {data} failed.", exc_info=exc)
            # TODO send an error message to the receiver?

    # Continuous listening and message handling
    def listen(self, stop_event: Event = SimpleEvent(), waiting_time: int = 100, **kwargs) -> None:
        """Listen for zmq communication until `stop_event` is set or until KeyboardInterrupt.

        :param stop_event: Event to stop the listening loop.
        :param waiting_time: Time to wait for a readout signal in ms.
        """
        self.stop_event = stop_event
        poller = self._listen_setup(**kwargs)
        # Loop
        try:
            while not stop_event.is_set():
                self._listen_loop_element(poller=poller, waiting_time=waiting_time)
        except KeyboardInterrupt:
            pass  # User stops the loop
        finally:
            # Close
            self._listen_close(waiting_time=waiting_time)

    def _listen_setup(self) -> zmq.Poller:
        """Setup for listening.

        If you add your own sockets, remember to poll only for incoming messages.
        """
        self.log.info(f"Start to listen as '{self.name}'.")
        # Prepare
        poller = zmq.Poller()
        poller.register(self.socket, zmq.POLLIN)

        # open communication
        self.sign_in()
        self.next_beat = time.perf_counter() + heartbeat_interval
        return poller

    def _listen_loop_element(
        self, poller: zmq.Poller, waiting_time: Optional[int]
    ) -> dict[zmq.Socket, int]:
        """Check the socks for incoming messages and handle them.

        :param waiting_time: Timeout of the poller in ms.
        """
        socks = dict(poller.poll(waiting_time))
        if self.socket in socks:
            self.read_and_handle_message()
            del socks[self.socket]
        elif (now := time.perf_counter()) > self.next_beat:
            self.heartbeat()
            self.next_beat = now + heartbeat_interval
        return socks

    def _listen_close(self, waiting_time: Optional[int] = None) -> None:
        """Close the listening loop."""
        self.log.info(f"Stop listen as '{self.name}'.")
        self.sign_out()

    # Message handling in loop
    def read_and_handle_message(self) -> None:
        """Interpret incoming message, which have not been requested."""
        try:
            message = self.read_message(timeout=0)
        except (TimeoutError, JSONRPCError):
            # only responses / errors arrived.
            return
        self.log.debug(f"Handling message {message}")
        if not message.payload:
            return  # no payload, that means just a heartbeat
        self.handle_message(message=message)

    def handle_message(self, message: Message) -> None:
        if message.header_elements.message_type == MessageTypes.JSON:
            self.handle_json_message(message=message)
        else:
            self.handle_unknown_message_type(message=message)

    def handle_json_message(self, message: Message) -> None:
        try:
            data: dict[str, Any] = message.data  # type: ignore
        except JSONDecodeError as exc:
            self.log.exception(f"Could not decode json message {message}", exc_info=exc)
            return
        content = get_json_content_type(data)
        if JsonContentTypes.REQUEST in content:
            self.handle_json_request(message=message)
        elif JsonContentTypes.ERROR in content:
            self.handle_json_error(message=message)
        elif JsonContentTypes.RESULT in content:
            self.handle_json_result(message)
        else:
            self.log.error(f"Invalid JSON message received: {message}")

    def handle_json_request(self, message: Message) -> None:
        response = self.process_json_message(message=message)
        self.send_message(response)

    def process_json_message(self, message: Message) -> Message:
        self.current_message = message
        self.additional_response_payload = None
        self.log.info(f"Handling commands of {message}.")
        reply = self.rpc.process_request(message.payload[0])
        response = Message(
            message.sender,
            conversation_id=message.conversation_id,
            message_type=MessageTypes.JSON,
            data=reply,
            additional_payload=self.additional_response_payload
        )
        return response

    def handle_json_error(self, message: Message) -> None:
        self.log.warning(f"Error message from {message.sender!r} received: {message}")

    def handle_json_result(self, message: Message) -> None:
        self.log.warning(f"Unsolicited message from {message.sender!r} received: '{message}'")

    def handle_unknown_message_type(self, message: Message) -> None:
        self.log.warning(
            f"Message from {message.sender!r} with unknown message type "
            f"{message.header_elements.message_type} received: '{message.data}', "
            f"{message.payload!r}."
        )

    # Methods offered via RPC
    def set_log_level(self, level: str) -> None:
        """Set the log level."""
        plevel = PythonLogLevels[level]
        self.root_logger.setLevel(plevel)

    def shut_down(self) -> None:
        self.stop_event.set()
